# -*- coding: utf-8 -*-
from django.contrib import admin
from django.conf import settings

import reversion

from nodeshot.core.base.admin import BaseStackedInline, PublishActionsAdminMixin
from .models import Page, MenuItem


class PageAdmin(PublishActionsAdminMixin, reversion.VersionAdmin):
    list_display = ('title', 'slug', 'is_published', 'access_level', 'added', 'updated')
    list_display_links = ('title', 'slug')
    list_filter = ('is_published',)
    prepopulated_fields = { 'slug': ('title',) }
    readonly_fields = ("added", "updated")
    save_on_top = True
    search_fields = ['title', 'content']
    change_list_template = 'reversion_and_smuggler/change_list.html'
    ordering = ('id',)

    fieldsets = (
        (None, {
            'fields': ('title', 'slug', 'content')
        }),
        ('Settings', {
            # TODO: access_level conditional depending on NODESHOT_ACL_EDITABLE_DEFAULT and NODESHOT_ACL_CMS_PAGE_EDITABLE
            'fields': ('access_level', 'is_published', 'added', 'updated')
        }),
        ('Meta tags (optional)', {
            'fields': ('meta_description', 'meta_keywords', 'meta_robots')
        }),
    )

    if 'grappelli' in settings.INSTALLED_APPS:
        class Media:
            js = [
                '%sgrappelli/tinymce/jscripts/tiny_mce/tiny_mce.js' % settings.STATIC_URL,
                '%sgrappelli/tinymce_setup/tinymce_setup_ns.js' % settings.STATIC_URL,
            ]

        # enable editor for "content" field only
        def formfield_for_dbfield(self, db_field, **kwargs):
            field = super(PageAdmin, self).formfield_for_dbfield(db_field, **kwargs)

            if db_field.name == 'content':
                field.widget.attrs['class'] = 'html-editor %s' % field.widget.attrs.get('class', '')

            return field


class MenuItemInline(BaseStackedInline):
    model = MenuItem

    if 'grappelli' in settings.INSTALLED_APPS:
        sortable_field_name = 'order'
        classes = ('grp-collapse grp-open', )


class MenuItemAdmin(PublishActionsAdminMixin, reversion.VersionAdmin):
    list_display = (
        'name', 'url', 'classes',
        'order', 'is_published', 'access_level',
        'added', 'updated'
    )
    list_editable = ('order',)
    list_filter = ('is_published',)
    readonly_fields = ('added', 'updated')
    save_on_top = True
    change_list_template = 'reversion_and_smuggler/change_list.html'
    inlines = [MenuItemInline]

    def get_queryset(self, request):
        return MenuItem.objects.filter(parent=None)

    fieldsets = (
        (None, {
            'fields': ('name', 'url', 'classes')
        }),
        ('Settings', {
            'fields': ('access_level', 'is_published', 'order', 'added', 'updated')
        }),
    )

admin.site.register(Page, PageAdmin)
admin.site.register(MenuItem, MenuItemAdmin)



from django.core.urlresolvers import NoReverseMatch

from rest_framework import serializers
from rest_framework.fields import Field
from rest_framework.reverse import reverse


class ExtraFieldSerializerOptions(serializers.ModelSerializerOptions):
    """
    Meta class options for ExtraFieldSerializerOptions
    """
    def __init__(self, meta):
        super(ExtraFieldSerializerOptions, self).__init__(meta)
        self.non_native_fields = getattr(meta, 'non_native_fields', ())


# TODO: rename / remove
class ExtraFieldSerializer(serializers.ModelSerializer):
    """
    ModelSerializer in which non native extra fields can be specified.
    """

    _options_class = ExtraFieldSerializerOptions

    def restore_object(self, attrs, instance=None):
        """
        Deserialize a dictionary of attributes into an object instance.
        You should override this method to control how deserialized objects
        are instantiated.
        """
        for field in self.opts.non_native_fields:
            attrs.pop(field)

        return super(ExtraFieldSerializer, self).restore_object(attrs, instance)

    def to_native(self, obj):
        """
        Serialize objects -> primitives.
        """
        ret = self._dict_class()
        ret.fields = self._dict_class()

        for field_name, field in self.fields.items():
            if field.read_only and obj is None:
                continue
            field.initialize(parent=self, field_name=field_name)
            key = self.get_field_key(field_name)

            # skips to next iteration but permits to show the field in API browser
            try:
                value = field.field_to_native(obj, field_name)
            except AttributeError as e:
                if field_name in self.opts.non_native_fields:
                    continue
                else:
                    raise AttributeError(e.message)

            method = getattr(self, 'transform_%s' % field_name, None)
            if callable(method):
                value = method(obj, value)
            if not getattr(field, 'write_only', False):
                ret[key] = value
            ret.fields[key] = self.augment_field(field, field_name, key, value)

        return ret


class DynamicRelationshipsMixin(object):
    """
    Django Rest Framework Serializer Mixin
    which adds the possibility to dynamically add relationships to a serializer.

    To add a relationship, use the class method "add_relationship", this way:

    >>> SerializerName.add_relationship('relationship_name', 'view_name', 'lookup_field')

    for example:

    >>> from nodeshot.core.nodes.serializers import NodeDetailSerializer
    >>> NodeDetailSerializer.add_relationship(**{
        'name': 'comments',
        'view_name': 'api_node_comments',
        'lookup_field': 'slug'
    })
    """
    _relationships = {}

    @classmethod
    def add_relationship(_class, name,
                         view_name=None, lookup_field=None,
                         serializer=None, many=False, queryset=None,
                         function=None):
        """ adds a relationship to serializer
        :param name: relationship name (dictionary key)
        :type name: str
        :param view_name: view name as specified in urls.py
        :type view_name: str
        :param lookup_field: lookup field, usually slug or id/pk
        :type lookup_field: str
        :param serializer: Serializer class to use for relationship
        :type serializer: Serializer
        :param many: indicates if it's a list or a single element, defaults to False
        :type many: bool
        :param queryset: queryset string representation to use for the serializer
        :type queryset: QuerySet
        :param function: function that returns the value to display (dict, list or str)
        :type function: function(obj, request)
        :returns: None
        """
        if view_name is not None and lookup_field is not None:
            _class._relationships[name] = {
                'type': 'link',
                'view_name': view_name,
                'lookup_field': lookup_field
            }
        elif serializer is not None and queryset is not None:
            _class._relationships[name] = {
                'type': 'serializer',
                'serializer': serializer,
                'many': many,
                'queryset': queryset
            }
        elif function is not None:
            _class._relationships[name] = {
                'type': 'function',
                'function': function
            }
        else:
            raise ValueError('missing arguments, either pass view_name and lookup_field or serializer and queryset')

    def get_lookup_value(self, obj, string):
        if '.' in string:
            if '()' in string:
                string = string.replace('()', '')
                is_method = True
            else:
                is_method = False
            levels = string.split('.')
            value = getattr(obj, levels.pop(0))
            if value is not None:
                for level in levels:
                    value = getattr(value, level)
                if is_method:
                    return value()
                else:
                    return value
            else:
                return None
        else:
            return getattr(obj, string)

    def get_relationships(self, obj):
        request = self.context['request']
        format = self.context['format']
        relationships = {}

        # loop over private _relationship attribute
        for key, options in self._relationships.iteritems():
            # if relationship is a link
            if options['type'] == 'link':
                # get lookup value
                lookup_value = self.get_lookup_value(obj, options['lookup_field'])
                # get URL
                value = reverse(options['view_name'],
                                args=[lookup_value],
                                request=request,
                                format=format)
            # if relationship is a serializer
            elif options['type'] == 'serializer':
                queryset = options['queryset'](obj, request)
                # get serializer representation
                value = options['serializer'](instance=queryset,
                                              context=self.context,
                                              many=options['many']).data
            elif options['type'] == 'function':
                value = options['function'](obj, request)
            else:
                raise ValueError('type %s not recognized' % options['type'])
            # populate new dictionary with value
            relationships[key] = value
        return relationships


class HyperlinkedField(Field):
    """
    Represents the instance, or a property on the instance, using hyperlinking.
    """
    read_only = True

    def __init__(self, *args, **kwargs):
        self.view_name = kwargs.pop('view_name', None)
        # Optionally the format of the target hyperlink may be specified
        self.format = kwargs.pop('format', None)
        # Optionally specify arguments
        self.view_args = kwargs.pop('view_args', None)

        super(HyperlinkedField, self).__init__(*args, **kwargs)

    def field_to_native(self, obj, field_name):
        request = self.context.get('request', None)
        format = self.context.get('format', None)
        view_name = self.view_name

        # By default use whatever format is given for the current context
        # unless the target is a different type to the source.
        if format and self.format and self.format != format:
            format = self.format

        try:
            return reverse(view_name, args=self.view_args, request=request, format=format)
        except NoReverseMatch:
            pass

        raise Exception('Could not resolve URL for field using view name "%s"' % view_name)


class GeoJSONDefaultObjectSerializer(serializers.Field):
    """
    If no object serializer is specified, then this serializer will be applied
    as the default.
    """

    def __init__(self, source=None, context=None):
        # Note: Swallow context kwarg - only required for eg. ModelSerializer.
        super(GeoJSONDefaultObjectSerializer, self).__init__(source=source)


class GeoJSONPaginationSerializerOptions(serializers.SerializerOptions):
    """
    An object that stores the options that may be provided to a
    pagination serializer by using the inner `Meta` class.

    Accessible on the instance as `serializer.opts`.
    """
    def __init__(self, meta):
        super(GeoJSONPaginationSerializerOptions, self).__init__(meta)
        self.object_serializer_class = getattr(meta, 'object_serializer_class',
                                               GeoJSONDefaultObjectSerializer)


class GeoJSONBasePaginationSerializer(serializers.Serializer):
    """
    A custom class for geojson serializers.
    """
    _options_class = GeoJSONPaginationSerializerOptions
    results_field = 'features'

    def __init__(self, *args, **kwargs):
        """
        Override init to add in the object serializer field on-the-fly.
        """
        super(GeoJSONBasePaginationSerializer, self).__init__(*args, **kwargs)
        results_field = self.results_field
        object_serializer = self.opts.object_serializer_class
        if 'context' in kwargs:
            context_kwarg = {'context': kwargs['context']}
        else:
            context_kwarg = {}

        self.fields[results_field] = object_serializer(source='object_list', **context_kwarg)


class GeoJSONPaginationSerializer(GeoJSONBasePaginationSerializer):
    """
    A geoJSON implementation of a pagination serializer.
    """
    type = serializers.SerializerMethodField('get_type')

    def get_type(self, obj):
        """ returns FeatureCollection type for geojson """
        return "FeatureCollection"



from netfields.managers import NetQuery, NetWhere, NetManager
from nodeshot.core.base.managers import ExtendedManagerMixin, ACLMixin, AccessLevelQuerySet


class NetAccessLevelManager(NetManager, ExtendedManagerMixin, ACLMixin):
    """ NetManager + AccessLevelManager """

    def get_query_set(self):
        q = NetQuery(self.model, NetWhere)
        return AccessLevelQuerySet(self.model, using=self._db, query=q)


import simplejson as json

from django.contrib.gis.geos import GEOSGeometry, Point
from django.core.urlresolvers import reverse
from django.core.exceptions import ValidationError
from django.conf import settings
from django.contrib.auth import get_user_model
User = get_user_model()

from nodeshot.core.base.tests import BaseTestCase
from nodeshot.core.base.tests import user_fixtures
from nodeshot.core.nodes.models import Node

from .models import *


class NetTest(BaseTestCase):
    """
    Network Model Tests
    """
    
    fixtures = [
        'initial_data.json',
        user_fixtures,
        'test_layers.json',
        'test_status.json',
        'test_nodes.json',
        'test_routing_protocols.json',
        'test_devices.json',
        'test_interfaces.json',
        'test_ip_addresses.json'
    ]
    
    def test_ip_automatic_protocol_detection(self):
        """ automatic protocol detection """
        ip = Ip.objects.get(pk=1)
        ip.save()
        self.assertEquals('ipv4', ip.protocol)
        
        ip = Ip.objects.get(pk=2)
        ip.save()
        self.assertEquals('ipv6', ip.protocol)
    
    def test_interface_data_ip_addresses(self):
        ip = Ip.objects.get(pk=1)
        ip.save()
        
        ip_addresses = ip.interface.data['ip_addresses'].replace(' ', '').split(',')
        self.assertEqual(len(ip_addresses), ip.interface.ip_set.count())
        
        for ip in ip.interface.ip_set.all():
            self.assertIn(str(ip.address), ip_addresses)
    
    def test_device_manager(self):
        self.assertEqual(
            list(Device.objects.access_level_up_to('public').filter(location__distance_gte=(Point(41, 12), 8000))),
            list(Device.objects.filter(access_level__lte=0, location__distance_gte=(Point(41, 12), 8000)))
        )
    
    def test_netacl_manager(self):
        self.assertEqual(
            list(Ip.objects.access_level_up_to('public').filter(address__net_contained='172.16.40.0/24')),
            list(Ip.objects.filter(access_level__lte=0, address__net_contained='172.16.40.0/24'))
        )
    
    def test_device_inherits_node_location(self):
        """ ensure device location defaults to node location if empty """
        d = Device.objects.create(name='test_device_location', node_id=1, type='radio')
        self.assertEqual(d.location, d.node.geometry)
    
    def test_device_inherits_node_point(self):
        """ ensure device location defaults to node location if empty """
        node = Node.objects.get(pk=1)
        node.geometry = GEOSGeometry("POLYGON((12.501664825436 41.900427664574,12.524409957883 41.897552804356,12.53925866699 41.886499358789,12.495828338623 41.871289758828,12.478318878173 41.891611016451,12.502179809565 41.900810969491,12.501664825436 41.900427664574))")
        node.save()
        d = Device.objects.create(name='test_device_location', node_id=1, type='radio')
        self.assertEqual(d.location, d.node.point)
    
    def test_device_inherits_node_elevation(self):
        """ ensure device elevation is inherithed from node """
        d = Device.objects.create(name='test_device_location', node_id=1, type='radio')
        self.assertEqual(d.elev, d.node.elev)
    
    def test_shortcuts(self):
        d = Device.objects.create(name='test_device_location', node_id=1, type='radio')
        self.assertEqual(d.shortcuts['user'], d.node.user)
        self.assertEqual(d.owner, d.node.user)
        self.assertEqual(d.shortcuts['layer'], d.node.layer)
        self.assertEqual(d.layer, d.node.layer)
    
    def test_device_list_api(self):
        """ API device list """
        url = reverse('api_device_list')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['results']), Device.objects.access_level_up_to('public').count())
    
    def test_device_list_search_api(self):
        """ API device list search """
        url = reverse('api_device_list')
        response = self.client.get(url, { 'search': 'RDP' })
        self.assertEqual(len(response.data['results']), 1)
    
    def test_device_list_api_ACL(self):
        """ API device list ACL """
        device = Device.objects.get(pk=1)
        device.access_level = 2
        device.save()
        
        url = reverse('api_device_list')
        response = self.client.get(url)
        self.assertEqual(len(response.data['results']), Device.objects.access_level_up_to('public').count())
        
        self.client.login(username='admin', password='tester')
        response = self.client.get(url)
        user_1 = User.objects.get(pk=1)
        self.assertEqual(len(response.data['results']), Device.objects.accessible_to(user_1).count())
    
    def test_device_details_api(self):
        url = reverse('api_device_details', args=[1])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
    
    def test_device_details_api_permissions(self):
        """ device detail permissions: only owner or admins can alter data """
        # non owner can only read
        url = reverse('api_device_details', args=[1])
        response = self.client.patch(url, { 'description': 'permission test' })
        self.assertEqual(response.status_code, 403)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 403)
        
        # login as non-owner
        self.client.login(username='pisano', password='tester')
        response = self.client.patch(url, { 'description': 'permission test' })
        self.assertEqual(response.status_code, 403)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 403)
        
        # login as owner
        device = Device.objects.get(pk=1)
        self.client.login(username=device.owner.username, password='tester')
        
        # owner can edit
        response = self.client.patch(url, { 'description': 'permission test' })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['description'], 'permission test')
        # check DB
        device = Device.objects.get(pk=1)
        self.assertEqual(device.description, 'permission test')
        
        # admin can edit too
        self.client.logout()
        self.client.login(username='admin', password='tester')
        response = self.client.patch(url, { 'description': 'i am the admin' })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['description'], 'i am the admin')
        # check DB
        device = Device.objects.get(pk=1)
        self.assertEqual(device.description, 'i am the admin')
    
    def test_node_device_list_api(self):
        """ API node device list """
        
        url = reverse('api_node_devices', args=[Node.objects.get(pk=7).slug])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['results']), 2)
        
        url = reverse('api_node_devices', args=['idontexist'])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
    
    def test_node_device_list_api_ACL(self):
        """ API device list ACL """
        device = Device.objects.get(pk=1)
        device.access_level = 2
        device.save()
        
        node = Node.objects.get(pk=7)
        url = reverse('api_node_devices', args=[node.slug])
        response = self.client.get(url)
        self.assertEqual(len(response.data['results']), Device.objects.filter(node_id=node.id).access_level_up_to('public').count())
        
        self.client.login(username='admin', password='tester')
        response = self.client.get(url)
        user_1 = User.objects.get(pk=1)
        self.assertEqual(len(response.data['results']), Device.objects.filter(node_id=node.id).accessible_to(user_1).count())
    
    def test_device_creation_api_permissions(self):
        """ create device permissions: only node owner or admins can add devices """
        # non owner can only read
        node = Node.objects.get(pk=7)
        url = reverse('api_node_devices', args=[node.slug])
        
        device_data = {
            "name": "device creation test",
            "type": "radio",
            "description": "device creation test",
            "data": { "is_unittest": "true" }
        }
        
        # unauthenticated can't create
        response = self.client.post(url, device_data)
        self.assertEqual(response.status_code, 403)
        
        # login as non-owner
        self.client.login(username='pisano', password='tester')
        response = self.client.post(url, json.dumps(device_data), content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        
        # login as node owner
        self.client.login(username=node.owner.username, password='tester')
        
        # node owner can create device
        response = self.client.post(url, json.dumps(device_data), content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['description'], 'device creation test')
        self.assertEqual(response.data['data']['is_unittest'], 'true')
        # check DB
        device = Device.objects.order_by('-id').all()[0]
        self.assertEqual(device.description, 'device creation test')
        self.assertEqual(device.data['is_unittest'], 'true')
        device.delete()
        
        # admin can create device too
        self.client.logout()
        self.client.login(username='admin', password='tester')
        device_data["data"] = json.dumps(device_data["data"])
        response = self.client.post(url, device_data)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['description'], 'device creation test')
        self.assertEqual(response.data['data']['is_unittest'], 'true')
        # check DB
        device = Device.objects.order_by('-id').all()[0]
        self.assertEqual(device.description, 'device creation test')
        self.assertEqual(device.data['is_unittest'], 'true')
    
    def test_interface_shortcuts(self):
        ethernet = Ethernet.objects.create(
            device_id=1,
            name='eth0',
            mac='00:27:22:38:13:f4',
            standard='fast',
            duplex='full'
        )
        self.assertEqual(ethernet.shortcuts['user'], ethernet.device.node.user)
        self.assertEqual(ethernet.owner, ethernet.device.node.user)
        self.assertEqual(ethernet.shortcuts['layer'], ethernet.device.node.layer)
        self.assertEqual(ethernet.layer, ethernet.device.node.layer)
        self.assertEqual(ethernet.shortcuts['node'], ethernet.device.node)
        self.assertEqual(ethernet.node, ethernet.device.node)
    
    def test_device_ethernet_api(self):
        """ API device ethernet interfaces """
        
        url = reverse('api_device_ethernet', args=[1])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)
        
        url = reverse('api_device_ethernet', args=[99])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
    
    def test_mac_address_rest_field(self):
        data = {
            "name": "eth0",
            "mac": "IAM WRONG",
            "tx_rate": 10000, 
            "rx_rate": 10000, 
            "data": {
                "status": "active"
            },
            "standard": "fast", 
            "duplex": "full"
        }
        self.client.login(username='romano', password='tester')
        url = reverse('api_device_ethernet', args=[2])
        
        response = self.client.post(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['mac'][0], 'Invalid mac address')
        
        data['mac'] = '00:27:22:38:13:f4'
        response = self.client.post(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 201)
    
    def test_device_ethernet_api_permissions(self):
        """ API device ethernet interfaces permissions to create """
        data = {
            "name": "eth0",
            "mac": "00:27:22:38:13:f4",
            "tx_rate": 10000, 
            "rx_rate": 10000, 
            "standard": "fast", 
            "duplex": "full",
            "data": { "status": "active" }
        }
        json_data = json.dumps(data)
        eth_count = Ethernet.objects.filter(device_id=2).count()
        url = reverse('api_device_ethernet', args=[2])
        
        # unlogged can't create
        response = self.client.post(url, json_data, content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(Ethernet.objects.filter(device_id=2).count(), eth_count)
        
        # non-owner can't create
        self.client.login(username='registered', password='tester')
        response = self.client.post(url, json_data, content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(Ethernet.objects.filter(device_id=2).count(), eth_count)
        self.client.logout()
        
        # owner can create
        self.client.login(username='romano', password='tester')
        response = self.client.post(url, json_data, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Ethernet.objects.filter(device_id=2).count(), eth_count+1)
        self.client.logout()
        Ethernet.objects.get(mac='00:27:22:38:13:f4').delete()
        
        # admin can create too
        self.client.login(username='admin', password='tester')
        response = self.client.post(url, json_data, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Ethernet.objects.filter(device_id=2).count(), eth_count+1)
    
    def test_device_ethernet_api_ACL(self):
        """ API device ethernet ACL """
        ethernet = Ethernet.objects.create(
            device_id=1,
            access_level=1,
            name='private',
            mac='00:27:22:38:13:a4',
            standard='fast',
            duplex='full'
        )
        
        url = reverse('api_device_ethernet', args=[1])
        device_url = reverse('api_device_details', args=[1])
        
        # unauthenticated can see only 1 record
        response = self.client.get(url)
        self.assertEqual(len(response.data), 1)
        # check device detail url too
        response = self.client.get(device_url)
        self.assertEqual(len(response.data['ethernet']), 1)
        
        # registered can see two
        self.client.login(username='registered', password='tester')
        response = self.client.get(url)
        self.assertEqual(len(response.data), 2)
        # check device detail url too
        response = self.client.get(device_url)
        self.assertEqual(len(response.data['ethernet']), 2)
        
        # set to higher access level
        ethernet.access_level = 2
        ethernet.save()
        
        # now registed can see only 1
        response = self.client.get(url)
        self.assertEqual(len(response.data), 1)
        # check device detail url too
        response = self.client.get(device_url)
        self.assertEqual(len(response.data['ethernet']), 1)
        
        # community can see two
        self.client.logout()
        self.client.login(username='community', password='tester')
        response = self.client.get(url)
        self.assertEqual(len(response.data), 2)
        # check device detail url too
        response = self.client.get(device_url)
        self.assertEqual(len(response.data['ethernet']), 2)
    
    def test_ethernet_details_api_permissions(self):
        """ ethernet details permissions: only owner or admins can alter data """
        # non owner can only read
        url = reverse('api_ethernet_details', args=[1])
        response = self.client.patch(url, { 'name': 'eth2' })
        self.assertEqual(response.status_code, 403)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 403)
        
        # login as non-owner
        self.client.login(username='pisano', password='tester')
        response = self.client.patch(url, { 'name': 'eth2' })
        self.assertEqual(response.status_code, 403)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 403)
        
        # login as owner
        ethernet = Ethernet.objects.get(pk=1)
        self.client.login(username=ethernet.owner.username, password='tester')
        
        # owner can edit
        response = self.client.patch(url, { 'name': 'eth2' })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['name'], 'eth2')
        # check DB
        ethernet = Ethernet.objects.get(pk=1)
        self.assertEqual(ethernet.name, 'eth2')
        
        # admin can edit too
        self.client.logout()
        self.client.login(username='admin', password='tester')
        response = self.client.patch(url, { 'name': 'admineth2' })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['name'], 'admineth2')
        # check DB
        ethernet = Ethernet.objects.get(pk=1)
        self.assertEqual(ethernet.name, 'admineth2')
    
    def test_ethernet_details_api_ACL(self):
        """ ethernet details ACL """
        # non owner can only read
        ethernet = Ethernet.objects.get(pk=1)
        ethernet.access_level = 1
        ethernet.save()
        url = reverse('api_ethernet_details', args=[1])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
        
        self.client.login(username='registered', password='tester')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
    
    def test_device_wireless_api(self):
        """ API device wireless interfaces """
        
        url = reverse('api_device_wireless', args=[1])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data), 1)
        
        url = reverse('api_device_wireless', args=[99])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
    
    def test_device_wireless_api_ACL(self):
        """ API device wireless ACL """
        wireless = Wireless.objects.create(
            device_id=1,
            access_level=1,
            name='private',
            mac='00:27:22:38:13:a5',
            standard='802.11n'
        )
        
        url = reverse('api_device_wireless', args=[1])
        device_url = reverse('api_device_details', args=[1])
        
        # unauthenticated can see 1
        response = self.client.get(url)
        self.assertEqual(len(response.data), 1)
        # check device detail url too
        response = self.client.get(device_url)
        self.assertEqual(len(response.data['wireless']), 1)
        
        # registered can see 2
        self.client.login(username='registered', password='tester')
        response = self.client.get(url)
        self.assertEqual(len(response.data), 2)
        # check device detail url too
        response = self.client.get(device_url)
        self.assertEqual(len(response.data['wireless']), 2)
        
        # set to higher access level
        wireless.access_level = 2
        wireless.save()
        
        # now registed can see only 1
        response = self.client.get(url)
        self.assertEqual(len(response.data), 1)
        # check device detail url too
        response = self.client.get(device_url)
        self.assertEqual(len(response.data['wireless']), 1)
        
        # community can see two
        self.client.logout()
        self.client.login(username='community', password='tester')
        response = self.client.get(url)
        self.assertEqual(len(response.data), 2)
        # check device detail url too
        response = self.client.get(device_url)
        self.assertEqual(len(response.data['wireless']), 2)
    
    def test_device_wireless_api_permissions(self):
        """ API device wireless interfaces permissions to create """
        data = {
            "name": "ath0",
            "mac": "00:27:22:38:13:c4",
            "tx_rate": 10000, 
            "rx_rate": 10000, 
            "standard": "802.11n",
            "data": { "status": "active" }
            
        }
        json_data = json.dumps(data)
        eth_count = Wireless.objects.filter(device_id=2).count()
        url = reverse('api_device_wireless', args=[2])
        
        # unlogged can't create
        response = self.client.post(url, json_data, content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(Wireless.objects.filter(device_id=2).count(), eth_count)
        
        # non-owner can't create
        self.client.login(username='registered', password='tester')
        response = self.client.post(url, json_data, content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.assertEqual(Wireless.objects.filter(device_id=2).count(), eth_count)
        self.client.logout()
        
        # owner can create
        self.client.login(username='romano', password='tester')
        response = self.client.post(url, json_data, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Wireless.objects.filter(device_id=2).count(), eth_count+1)
        self.client.logout()
        Wireless.objects.get(mac='00:27:22:38:13:c4').delete()
        
        # admin can create too
        self.client.login(username='admin', password='tester')
        response = self.client.post(url, json_data, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Wireless.objects.filter(device_id=2).count(), eth_count+1)
    
    def test_wireless_details_api_permissions(self):
        """ wireless details permissions: only owner or admins can alter data """
        # non owner can only read
        url = reverse('api_wireless_details', args=[2])
        response = self.client.patch(url, { 'name': 'eth2' })
        self.assertEqual(response.status_code, 403)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 403)
        
        # login as non-owner
        self.client.login(username='pisano', password='tester')
        response = self.client.patch(url, { 'name': 'eth2' })
        self.assertEqual(response.status_code, 403)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 403)
        
        # login as owner
        wireless = Wireless.objects.get(pk=2)
        self.client.login(username=wireless.owner.username, password='tester')
        
        # owner can edit
        response = self.client.patch(url, { 'name': 'eth2' })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['name'], 'eth2')
        # check DB
        wireless = Wireless.objects.get(pk=2)
        self.assertEqual(wireless.name, 'eth2')
        
        # admin can edit too
        self.client.logout()
        self.client.login(username='admin', password='tester')
        response = self.client.patch(url, { 'name': 'admineth2' })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['name'], 'admineth2')
        # check DB
        wireless = Wireless.objects.get(pk=2)
        self.assertEqual(wireless.name, 'admineth2')
    
    def test_wireless_details_api_ACL(self):
        """ wireless details ACL """
        # non owner can only read
        wireless = Wireless.objects.get(pk=2)
        wireless.access_level = 1
        wireless.save()
        url = reverse('api_wireless_details', args=[2])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
        
        self.client.login(username='registered', password='tester')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
    
    def test_interface_ip_address_get(self):
        # GET 404
        response = self.client.get(reverse('api_interface_ip', args=[99]))
        self.assertEqual(response.status_code, 404)
        
        # GET 200
        response = self.client.get(reverse('api_interface_ip', args=[1]))
        self.assertEqual(response.status_code, 200)
    
    def test_interface_ip_address_post_and_permissions(self):
        url = reverse('api_interface_ip', args=[1])
        data = {
            "address": "10.40.0.35",
            "netmask": "10.40.0.0/24"
        }
        
        # POST 403 unauthenticated
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 403)
        
        # POST 403 non owner can't add ip
        self.client.login(username='pisano', password='tester')
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        
        # POST 200 owner can add ip
        self.client.login(username=Interface.objects.find(1).owner.username, password='tester')
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 201)
        # ensure address
        ip = Ip.objects.last()
        self.assertEqual(ip.address.__str__(), '10.40.0.35')
        ip.delete()
        self.client.logout()
        
        # POST 200 also admin can add ip
        self.client.login(username='admin', password='tester')
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 201)
        # ensure address
        ip = Ip.objects.last()
        self.assertEqual(ip.address.__str__(), '10.40.0.35')
    
    def test_interface_ip_address_fields_validation(self):
        url = reverse('api_interface_ip', args=[1])
        data = {
            "address": "WRONG",
            "netmask": "VERY WRONG"
        }
        
        self.client.login(username=Interface.objects.find(1).owner.username, password='tester')
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['address'][0], 'Invalid ip address')
        self.assertEqual(response.data['netmask'][0], 'Invalid ip network')
        
        data = {
            "address": "10.40.0.1",
            "netmask": "10.40."
        }
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 400)
        
        data = {
            "address": "10.40.0.",
            "netmask": "10.40.0.0/24"
        }
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 400)
    
    def test_interface_ip_api_ACL(self):
        url = reverse('api_interface_ip', args=[1])
        
        # non owner can only read
        interface = Interface.objects.get(pk=1)
        ip_count = interface.ip_set.count()
        ip = interface.ip_set.first()
        ip.access_level = 1
        ip.save()
        response = self.client.get(url)
        self.assertEqual(len(response.data), interface.ip_set.access_level_up_to('public').count())
        self.assertEqual(len(response.data), ip_count-1)
        
        self.client.login(username='registered', password='tester')
        response = self.client.get(url)
        self.assertEqual(len(response.data), interface.ip_set.access_level_up_to('registered').count())
        self.assertEqual(len(response.data), ip_count)
    
    def test_ip_address_details_permissions(self):
        url = reverse('api_ip_details', args=[1])
        data = json.dumps({
            "address": "10.40.0.35",
            "netmask": "10.40.0.0/24"
        })
        
        # PUT 403 unauthenticated
        response = self.client.put(url, data, content_type='application/json')
        self.assertEqual(response.status_code, 403)
        
        # PUT 403 non owner can't edit ip
        self.client.login(username='pisano', password='tester')
        response = self.client.put(url, data, content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        
        # PUT 200 owner can edit ip
        self.client.login(username=Ip.objects.find(1).owner.username, password='tester')
        response = self.client.put(url, data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        # ensure address
        ip = Ip.objects.find(1)
        self.assertEqual(ip.address.__str__(), '10.40.0.35')
        self.client.logout()
        
        # PUT 200 also admin can edit ip
        self.client.login(username='admin', password='tester')
        data = json.dumps({
            "address": "10.40.0.36",
            "netmask": "10.40.0.0/24"
        })
        response = self.client.put(url, data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        # ensure address
        ip = Ip.objects.find(1)
        self.assertEqual(ip.address.__str__(), '10.40.0.36')
    
    def test_ip_details_api_ACL(self):
        # non owner can only read
        ip = Ip.objects.get(pk=1)
        ip.access_level = 1
        ip.save()
        url = reverse('api_ip_details', args=[1])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
        
        self.client.login(username='registered', password='tester')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
    
    def test_bridge_interfaces_validation(self):
        bridge = Bridge(device_id=1, mac='00:11:22:33:44:55', name='br0')
        bridge.save()
        
        # must bridge at least 2
        with self.assertRaises(ValidationError):
            bridge.interfaces.add(Interface.objects.find(1))
        
        bridge.interfaces.add(Interface.objects.find(1), Interface.objects.find(2))
        bridge.full_clean()
        
        # can't bridge different devices
        ethernet = Ethernet.objects.create(device_id=2, mac='00:00:22:33:44:55', name='br0')
        with self.assertRaises(ValidationError):
            bridge.interfaces.add(ethernet)
        
        # test add
        ethernet.device_id = 1
        ethernet.save()
        bridge.interfaces.add(ethernet)
        
        # can't bridge self
        with self.assertRaises(ValidationError):
            bridge.interfaces.add(bridge)
    
    def test_device_bridge_api(self):
        bridge = Bridge(device_id=1, mac='00:11:22:33:44:55', name='br0')
        bridge.save()
        bridge.interfaces.add(Interface.objects.find(1), Interface.objects.find(2))
        
        url = reverse('api_device_bridge', args=[1])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertIn('http://', response.data[0]['interfaces_links'][0])
        self.assertIn('http://', response.data[0]['interfaces_links'][1])
        bridge.delete()
        
        data = {
            "name": "br0test", 
            "mac": "00:11:22:33:44:55",
            "interfaces": []
        }
        
        # POST 403: unauthorized
        response = self.client.post(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.client.login(username='pisano', password='tester')
        response = self.client.post(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        
        # POST 400: missing interfaces
        self.client.login(username=bridge.owner.username, password='tester')
        response = self.client.post(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 400)
        
        data['interfaces'] = [1, 2]
        response = self.client.post(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 201)
        
        # POST 201
        bridge = Bridge.objects.last()
        self.assertEquals(bridge.name, 'br0test')
        self.assertEquals(bridge.mac, '00:11:22:33:44:55')
        self.assertEquals(bridge.interfaces.count(), 2)
        
        # POST 400: duplicate mac address
        response = self.client.post(url, json.dumps(data), content_type='application/json')
        self.assertEqual(response.status_code, 400)
    
    def test_bridge_details_api(self):
        bridge = Bridge(device_id=1, mac='00:11:22:33:44:55', name='br0')
        bridge.save()
        bridge.interfaces.add(Interface.objects.find(1), Interface.objects.find(2))
        
        url = reverse('api_bridge_details', args=[bridge.id])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertIn('http://', response.data['interfaces_links'][0])
        self.assertIn('http://', response.data['interfaces_links'][1])
        
        # PATCH 403 unauthenticated
        response = self.client.patch(url, {})
        self.assertEqual(response.status_code, 403)
        
        # PATCH 403 logged in as non owner
        self.client.login(username='pisano', password='tester')
        response = self.client.patch(url, {})
        self.assertEqual(response.status_code, 403)
        self.client.logout()
        
        # PATCH 200 logged in as owner
        self.client.login(username=bridge.owner.username, password='tester')
        response = self.client.patch(url, { 'name': 'brtest' })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['name'], 'brtest')
        
        # PATCH 400 empty interfaces
        response = self.client.patch(url, json.dumps({ 'interfaces': [] }), content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.client.logout()
        
        # PATCH 200 by admin
        self.client.login(username='admin', password='tester')
        response = self.client.patch(url, { 'name': 'admin' })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['name'], 'admin')
    
    # TODO: write tests for vlan, tunnel, vap


from celery import task
from django.utils.module_loading import import_by_path
from django.core import management


@task
def synchronize_external_layers(*args, **kwargs):
    """
    runs "python manage.py synchronize"
    """
    management.call_command('sync', *args, **kwargs)


# ------ Asynchronous tasks ------ #


@task
def push_changes_to_external_layers(node, external_layer, operation):
    """
    Sync other applications through their APIs by performing updates, adds or deletes.
    This method is designed to be performed asynchronously, avoiding blocking the user
    when he changes data on the local DB.

    :param node: the node which should be updated on the external layer.
    :type node: Node model instance
    :param operation: the operation to perform (add, change, delete)
    :type operation: string
    """
    # putting the model inside prevents circular imports
    # subsequent imports go and look into sys.modules before reimporting the module again
    # so performance is not affected
    from nodeshot.core.nodes.models import Node
    # get node because for some reason the node instance object is not passed entirely,
    # pheraphs because objects are serialized by celery or transport/queuing mechanism
    if not isinstance(node, basestring):
        node = Node.objects.get(pk=node.pk)
    # import synchronizer
    Synchronizer = import_by_path(external_layer.synchronizer_path)
    # create instance
    instance = Synchronizer(external_layer.layer)
    # call method only if supported
    if hasattr(instance, operation):
        getattr(instance, operation)(node)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from nodeshot.core.base.models import BaseDate


class Category(BaseDate):
    """
    Categories of services
    """
    name = models.CharField(_('name'), max_length=30)
    description = models.TextField(_('description'), blank=True, null=True)
    
    class Meta:
        app_label = 'services'
        db_table = 'service_category'
        verbose_name = _('category')
        verbose_name_plural = _('categories')
    
    def __unicode__(self):
        return '%s' % self.name


from django.contrib.gis.db import models
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ValidationError
from django.contrib.gis.measure import D
from django.contrib.gis.geos import Polygon, Point, GEOSException

from django_hstore.fields import DictionaryField

from nodeshot.core.base.models import BaseDate
from nodeshot.core.nodes.models import Node

from ..settings import settings, NODES_MINIMUM_DISTANCE, HSTORE_SCHEMA
from ..managers import LayerManager
from ..signals import layer_is_published_changed


class Layer(BaseDate):
    """
    Layer Model
    A layer represent a categorization of nodes.
    Layers might have geographical boundaries and might be managed by certain organizations.
    """
    name = models.CharField(_('name'), max_length=50, unique=True)
    slug = models.SlugField(max_length=50, db_index=True, unique=True)
    description = models.CharField(_('description'), max_length=250, blank=True, null=True,
                                   help_text=_('short description of this layer'))
    text = models.TextField(_('extended text'), blank=True, null=True,
                            help_text=_('extended description, specific instructions, links, ecc.'))
    # record management
    is_published = models.BooleanField(_('published'), default=True)
    is_external = models.BooleanField(_('is it external?'), default=False)
    # geographic related fields
    area = models.GeometryField(_('area'), help_text=_('If a polygon is used nodes of this layer will have to be contained in it.\
                                                        If a point is used nodes of this layer can be located anywhere. Lines are not allowed.'))
    # organizational
    organization = models.CharField(_('organization'), max_length=255, blank=True,
                                    help_text=_('Organization which is responsible to manage this layer'))
    website = models.URLField(_('Website'), blank=True, null=True)
    email = models.EmailField(_('email'), blank=True,
                              help_text=_("""possibly an email address that delivers messages to all the active participants;
                                          if you don't have such an email you can add specific users in the "mantainers" field"""))
    mantainers = models.ManyToManyField(settings.AUTH_USER_MODEL, verbose_name=_('mantainers'), blank=True,
                                        help_text=_('you can specify the users who are mantaining this layer so they will receive emails from the system'))
    # settings
    nodes_minimum_distance = models.IntegerField(default=NODES_MINIMUM_DISTANCE,
                                                 help_text=_('minimum distance between nodes in meters, 0 means there is no minimum distance'))
    new_nodes_allowed = models.BooleanField(_('new nodes allowed'), default=True, help_text=_('indicates whether users can add new nodes to this layer'))
    data = DictionaryField(_('extra data'), schema=HSTORE_SCHEMA, null=True, editable=False)

    # default manager
    objects = LayerManager()

    # this is needed to check if the is_published is changing
    # explained here:
    # http://stackoverflow.com/questions/1355150/django-when-saving-how-can-you-check-if-a-field-has-changed
    _current_is_published = None

    class Meta:
        db_table = 'layers_layer'
        app_label = 'layers'

    def __unicode__(self):
        return self.name

    def __init__(self, *args, **kwargs):
        """ Fill _current_is_published """
        super(Layer, self).__init__(*args, **kwargs)
        # set current is_published, but only if it is an existing layer
        if self.pk:
            self._current_is_published = self.is_published

    def save(self, *args, **kwargs):
        """
        intercepts changes to is_published and fires layer_is_published_changed signal
        """
        super(Layer, self).save(*args, **kwargs)

        # if is_published of an existing layer changes
        if self.pk and self.is_published != self._current_is_published:
            # send django signal
            layer_is_published_changed.send(
                sender=self.__class__,
                instance=self,
                old_is_published=self._current_is_published,
                new_is_published=self.is_published
            )
            # unpublish nodes
            self.update_nodes_published()

        # update _current_is_published
        self._current_is_published = self.is_published

    def clean(self):
        """
        Ensure area is either a Point or a Polygon
        """
        if not isinstance(self.area, (Polygon, Point)):
            raise ValidationError('area can be only of type Polygon or Point')

    @property
    def center(self):
        # if area is point just return that
        if isinstance(self.area, Point) or self.area is None:
            return self.area
        # otherwise return point_on_surface or centroid
        try:
            # point_on_surface guarantees that the point is within the geometry
            return self.area.point_on_surface
        except GEOSException:
            # fall back on centroid which may not be within the geometry
            # for example, a horseshoe shaped polygon
            return self.area.centroid

    def update_nodes_published(self):
        """ publish or unpublish nodes of current layer """
        if self.pk:
            self.node_set.all().update(is_published=self.is_published)

    if 'grappelli' in settings.INSTALLED_APPS:
        @staticmethod
        def autocomplete_search_fields():
            return ('name__icontains', 'slug__icontains')


# ------ Add Layer related methods to Node class ------ #


@property
def intersecting_layers(self):
    return Layer.objects.filter(area__contains=self.point)

Node.intersecting_layers = intersecting_layers


# ------ Additional validation for Node model ------ #


def new_nodes_allowed_for_layer(self):
    """
    ensure new nodes are allowed for this layer
    """
    if not self.pk and self.layer and not self.layer.new_nodes_allowed:
        raise ValidationError(_('New nodes are not allowed for this layer'))


def nodes_minimum_distance_validation(self):
    """
    if minimum distance is specified, ensure node is not too close to other nodes;
    """
    if self.layer and self.layer.nodes_minimum_distance:
        minimum_distance = self.layer.nodes_minimum_distance
        # TODO - lower priority: do this check only when coordinates are changing
        near_nodes = Node.objects.exclude(pk=self.id).filter(geometry__distance_lte=(self.geometry, D(m=minimum_distance))).count()
        if near_nodes > 0:
            raise ValidationError(_('Distance between nodes cannot be less than %s meters') % minimum_distance)


def node_contained_in_layer_area_validation(self):
    """
    if layer defines an area, ensure node coordinates are contained in the area
    """
    # if area is a polygon ensure it contains the node
    if self.layer and isinstance(self.layer.area, Polygon) and not self.layer.area.contains(self.geometry):
        raise ValidationError(_('Node must be inside layer area'))


Node.add_validation_method(new_nodes_allowed_for_layer)
Node.add_validation_method(nodes_minimum_distance_validation)
Node.add_validation_method(node_contained_in_layer_area_validation)



"""
Fixed Github Backend for social auth
"""

from social_auth.backends.contrib.github import GithubBackend as BaseGithubBackend,\
                                                GithubAuth


class GithubBackend(BaseGithubBackend):
    """Github OAuth authentication backend"""

    def get_user_details(self, response):
        """Return user details from Github account"""
        name = response.get('name') or ''
        details = {'username': response.get('login')}

        try:
            email = self._fetch_emails(response.get('access_token'))[0]
            if isinstance(email, dict):
                email = email['email']
        except IndexError:
            details['email'] = ''
        else:
            details['email'] = email

        try:
            # GitHub doesn't separate first and last names. Let's try.
            first_name, last_name = name.split(' ', 1)
        except ValueError:
            details['first_name'] = name
        else:
            details['first_name'] = first_name
            details['last_name'] = last_name
        return details


# Backend definition
BACKENDS = {
    'github': GithubAuth,
}



from django.conf.urls import patterns, url


urlpatterns = patterns('nodeshot.core.nodes.views',  # noqa
    url(r'^nodes/$', 'node_list', name='api_node_list'),
    url(r'^nodes.geojson$', 'geojson_list', name='api_node_gejson_list'),
    url(r'^nodes/(?P<slug>[-\w]+)/$', 'node_details', name='api_node_details'),

    # images
    url(r'^nodes/(?P<slug>[-\w]+)/images/$', 'node_images', name='api_node_images'),
    url(r'^nodes/(?P<slug>[-\w]+)/images/(?P<pk>[0-9]+)/$', 'node_image_detail', name='api_node_image_detail'),

    # status
    url(r'^status/$', 'status_list', name='api_status_list'),

    # utils
    url(r'^elevation/$', 'elevation_profile', name='api_elevation_profile'),
)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import DataMigration
from django.db import models

class Migration(DataMigration):

    OLD_STRING = "javascript:$.createModal({message:'not implemented yet'})"
    NEW_STRING = "#layers"

    def forwards(self, orm):
        "Write your forwards methods here."
        for menuitem in orm['cms.menuitem'].objects.filter(name__iexact='layers'):
            menuitem.url = self.NEW_STRING
            menuitem.save()

    def backwards(self, orm):
        "Write your backwards methods here."
        for menuitem in orm['cms.menuitem'].objects.filter(name__iexact='layers'):
            menuitem.url = self.OLD_STRING
            menuitem.save()

    models = {
        u'cms.menuitem': {
            'Meta': {'ordering': "['order']", 'object_name': 'MenuItem'},
            'access_level': ('django.db.models.fields.SmallIntegerField', [], {'default': '0'}),
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2014, 11, 28, 0, 0)'}),
            'classes': ('django.db.models.fields.CharField', [], {'max_length': '50', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_published': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'}),
            'order': ('django.db.models.fields.PositiveIntegerField', [], {'blank': 'True'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['cms.MenuItem']", 'null': 'True', 'blank': 'True'}),
            'updated': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2014, 11, 28, 0, 0)'}),
            'url': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        u'cms.page': {
            'Meta': {'object_name': 'Page'},
            'access_level': ('django.db.models.fields.SmallIntegerField', [], {'default': '0'}),
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2014, 11, 28, 0, 0)'}),
            'content': ('django.db.models.fields.TextField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_published': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'meta_description': ('django.db.models.fields.CharField', [], {'max_length': '255', 'blank': 'True'}),
            'meta_keywords': ('django.db.models.fields.CharField', [], {'max_length': '255', 'blank': 'True'}),
            'meta_robots': ('django.db.models.fields.CharField', [], {'default': "'index, follow'", 'max_length': '50'}),
            'slug': ('django.db.models.fields.SlugField', [], {'unique': 'True', 'max_length': '50', 'blank': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '50'}),
            'updated': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2014, 11, 28, 0, 0)'})
        }
    }

    complete_apps = ['cms']
    symmetrical = True



from django import forms
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth import get_user_model
User = get_user_model()

from .models import PasswordReset
from .settings import EMAIL_CONFIRMATION


__all__ = [
    'ResetPasswordForm',
    'ResetPasswordKeyForm'
]


class ResetPasswordForm(forms.Form):
    email = forms.EmailField(
        label=_("Email"),
        required=True,
        widget=forms.TextInput(attrs={"size": "30"})
    )

    def clean_email(self):
        """ ensure email is in the database """
        if EMAIL_CONFIRMATION:
            from .models import EmailAddress
            condition = EmailAddress.objects.filter(
                email__iexact=self.cleaned_data["email"],
                verified=True
            ).count() == 0
        else:
            condition = User.objects.get(
                email__iexact=self.cleaned_data["email"],
                is_active=True
            ).count() == 0

        if condition is True:
            raise forms.ValidationError(
                _("Email address not verified for any user account")
            )
        return self.cleaned_data["email"]

    def save(self, **kwargs):
        PasswordReset.objects.create_for_user(self.cleaned_dat["email"])
        return self.cleaned_data["email"]


class ResetPasswordKeyForm(forms.Form):
    password1 = forms.CharField(
        label=_("New Password"),
        widget=forms.PasswordInput(render_value=False)
    )
    password2 = forms.CharField(
        label=_("New Password (again)"),
        widget=forms.PasswordInput(render_value=False)
    )

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user", None)
        self.temp_key = kwargs.pop("temp_key", None)
        super(ResetPasswordKeyForm, self).__init__(*args, **kwargs)

    def clean_password2(self):
        if "password1" in self.cleaned_data and "password2" in self.cleaned_data:
            if self.cleaned_data["password1"] != self.cleaned_data["password2"]:
                raise forms.ValidationError(_("Password confirmation mismatch."))
        return self.cleaned_data["password2"]

    def save(self):
        # set the new user password
        user = self.user
        user.set_password(self.cleaned_data["password1"])
        user.save()
        # mark password reset object as reset
        PasswordReset.objects.filter(temp_key=self.temp_key).update(reset=True)



import os
from django.conf import settings


PUBLIC_PIPE = getattr(settings, 'NODESHOT_WEBSOCKETS_PUBLIC_PIPE', '%s/nodeshot.websockets.public' % os.path.dirname(settings.SITE_ROOT))
PRIVATE_PIPE = getattr(settings, 'NODESHOT_WEBSOCKETS_PRIVATE_PIPE', '%s/nodeshot.websockets.private' % os.path.dirname(settings.SITE_ROOT))
DOMAIN = settings.DOMAIN
PATH = getattr(settings, 'NODESHOT_WEBSOCKETS_PATH', '')
LISTENING_ADDRESS = getattr(settings, 'NODESHOT_WEBSOCKETS_LISTENING_ADDRESS', '0.0.0.0')
LISTENING_PORT = getattr(settings, 'NODESHOT_WEBSOCKETS_LISTENING_PORT', 8080)
REGISTER = getattr(settings, 'NODESHOT_WEBSOCKETS_REGISTER', (
    'nodeshot.core.websockets.registrars.nodes',
    'nodeshot.core.websockets.registrars.notifications',
))




LINK_STATUS = {
    'archived': -3,
    'disconnected': -2,
    'down': -1,
    'planned': 0,
    'testing': 1,
    'active': 2
}

LINK_TYPES = {
    'radio': 1,
    'ethernet': 2,
    'fiber': 3,
    'other_wired': 4,
    'virtual': 5
}

# TODO: pheraphs this would be better to make it customizable from settings
METRIC_TYPES = {
    'ETX': 'etx',
    'ETC': 'etc',
    'HOP': 'hop'
}



from django.contrib import admin
from django.conf import settings

from nodeshot.core.base.admin import BaseAdmin, BaseGeoAdmin, BaseStackedInline

from .models.choices import INTERFACE_TYPES
from .models import *


class DeviceInline(BaseStackedInline):
    model = Device
    
    if not settings.DEBUG:
        readonly_fields =  [
            'status', 'first_seen', 'last_seen'
        ] + BaseStackedInline.readonly_fields
    
    if 'grappelli' in settings.INSTALLED_APPS:
        classes = ('grp-collapse grp-open', )
        
        raw_id_fields = ('routing_protocols',)
        autocomplete_lookup_fields = {
            'm2m': ['routing_protocols']
        }


class EthernetInline(BaseStackedInline):
    model = Ethernet


class WirelessInline(BaseStackedInline):
    model = Wireless


class BridgeInline(BaseStackedInline):
    model = Bridge


class TunnelInline(BaseStackedInline):
    model = Tunnel


class VlanInline(BaseStackedInline):
    model = Vlan


class IpInline(BaseStackedInline):
    model = Ip


class VapInline(BaseStackedInline):
    model = Vap


class DeviceAdmin(BaseGeoAdmin):
    list_filter   = ('added', 'updated',)
    list_display  = ('name', 'status', 'node', 'type', 'access_level', 'added', 'updated')
    search_fields = ('name', 'type')
    inlines = [EthernetInline, WirelessInline, BridgeInline, TunnelInline, VlanInline]
    
    raw_id_fields = ('node', 'routing_protocols')
    autocomplete_lookup_fields = {
        'fk': ['node'],
        'm2m': ['routing_protocols']
    }
    
    exclude = ('shortcuts',)
    
    if not settings.DEBUG:
        readonly_fields =  [
            'status', 'first_seen', 'last_seen'
        ] + BaseAdmin.readonly_fields


class InterfaceAdmin(BaseAdmin):
    list_display  = ('mac', 'name', 'type', 'device', 'added', 'updated')
    search_fields = ('mac',)
    exclude = ('shortcuts',)
    inlines = (IpInline,)
    
    raw_id_fields = ('device',)
    autocomplete_lookup_fields = {
        'fk': ['device'],
    }


class WirelessAdmin(InterfaceAdmin):
    inlines = (IpInline, VapInline)   


class RoutingProtocolAdmin(BaseAdmin):
    list_display   = ('name', 'version')


class IpAdmin(BaseAdmin):
    list_display = ('address', 'netmask', 'protocol', 'added', 'updated')
    list_filter = ('protocol', 'added', 'updated')
    search_fields = ('address',)
    
    raw_id_fields = ('interface',)
    autocomplete_lookup_fields = {
        'fk': ['interface'],
    }
    
    readonly_fields = ['protocol'] + BaseAdmin.readonly_fields


from django import forms
from .models.interfaces.bridge import validate_bridged_interfaces

class BridgeForm(forms.ModelForm):
    class Meta:
        model = Bridge
    
    def __init__(self, *args, **kwargs):
        """ only interfaces of the same device can be bridged """
        super(BridgeForm, self).__init__(*args, **kwargs)
        self.fields['interfaces'].queryset = Interface.objects.filter(device_id=self.instance.device_id) \
                                             .exclude(type=INTERFACE_TYPES.get('bridge'))
    
    def clean_interfaces(self):
        """ interface many2many validation """
        interfaces = self.cleaned_data.get('interfaces', [])
        if interfaces:
            # custom signal
            validate_bridged_interfaces(
                sender=self.instance.interfaces,
                instance=self.instance,
                action="pre_add",
                reverse=False,
                model=self.instance.interfaces.model,
                pk_set=interfaces
            )
        return self.cleaned_data['interfaces']


class BridgeAdmin(InterfaceAdmin):
    form = BridgeForm 


admin.site.register(Device, DeviceAdmin)
admin.site.register(Interface, InterfaceAdmin)

admin.site.register(Ethernet, InterfaceAdmin)
admin.site.register(Wireless, WirelessAdmin)
admin.site.register(Bridge, BridgeAdmin)
admin.site.register(Tunnel, InterfaceAdmin)
admin.site.register(Vlan, InterfaceAdmin)

admin.site.register(RoutingProtocol, RoutingProtocolAdmin)
admin.site.register(Ip, IpAdmin)


# ------ Add Device Inlines to NodeAdmin ------ #

from nodeshot.core.nodes.admin import NodeAdmin

NodeAdmin.inlines = [DeviceInline] + NodeAdmin.inlines



from django.contrib import admin
from django.conf import settings

from nodeshot.core.base.admin import BaseAdmin, BaseStackedInline

from .models import Category, Service, ServiceLogin, Url


class CategoryAdmin(BaseAdmin):
    list_display  = ('name', 'description', 'added', 'updated')
    ordering = ('name',)
    search_fields = ('name', 'description')


class UrlInline(BaseStackedInline):
    model = Url
    
    if 'grappelli' in settings.INSTALLED_APPS:
        # define the raw_id_fields
        raw_id_fields = ('ip',)
        # define the autocomplete_lookup_fields
        autocomplete_lookup_fields = {
            'fk': ['ip'],
        }


class ServiceLoginInline(BaseStackedInline):
    model = ServiceLogin


class ServiceAdmin(BaseAdmin):
    list_display  = ('name', 'device', 'category', 'access_level', 'status', 'is_published', 'added', 'updated')
    list_filter   = ('category', 'status', 'is_published')
    search_fields = ('name', 'description', 'documentation_url')
    inlines = (UrlInline, ServiceLoginInline,)

    raw_id_fields = ('device',)
    autocomplete_lookup_fields = {
        'fk': ('device',),
    }
    
    # enable editor for "description" only
    html_editor_fields = ['description']

admin.site.register(Category, CategoryAdmin)
admin.site.register(Service, ServiceAdmin)



from django.conf import settings


API_PREFIX = getattr(settings, 'NODESHOT_API_PREFIX', 'api/v1/')
API_APPS_ENABLED = getattr(settings, 'NODESHOT_API_APPS_ENABLED', [
    'nodeshot.core.nodes',
    'nodeshot.core.layers',
    'nodeshot.core.cms',
    'nodeshot.community.profiles',
    'nodeshot.community.participation',
    'nodeshot.community.notifications',
    'nodeshot.community.mailing',
    'nodeshot.networking.net',
    'nodeshot.networking.links',
    'nodeshot.networking.services',
    'nodeshot.interop.open311',
])



# -*- coding: utf-8 -*-
"""
this app can import data from older nodeshot versions (0.9)
developed to import data from map.ninux.org into the new version of nodeshot
"""

from nodeshot.core.base.utils import check_dependencies

check_dependencies(
    dependencies=[
        'nodeshot.core.nodes',
        'nodeshot.core.layers',
        'nodeshot.networking.net',
        'nodeshot.networking.links',
        'nodeshot.community.mailing',
        'nodeshot.community.profiles',
    ],
    module='nodeshot.interop.oldimporter'
)

from django.db import models
from django.utils.translation import ugettext_lazy as _

from .choices import *


__all__ = [
    'OldNode',
    'OldDevice',
    'OldHna',
    'OldInterface',
    'OldLink',
    'OldStatistic',
    'OldContact',
    'OldUser'
]


class OldNode(models.Model):
    name = models.CharField(_('name'), max_length=50, unique=True)
    slug = models.SlugField(max_length=50, db_index=True, unique=True)
    owner = models.CharField(_('owner'), max_length=50, blank=True, null=True)
    description = models.CharField(_('description'), max_length=200, blank=True, null=True)
    postal_code = models.CharField(_('postal code'), max_length=10, blank=True)
    email = models.EmailField()
    email2 = models.EmailField(blank=True, null=True)
    email3 = models.EmailField(blank=True, null=True)
    password =  models.CharField(max_length=255, help_text=_('Use "[algo]$[salt]$[hexdigest]" or use the  <a href="password/">change password form</a>.'))
    lat = models.FloatField(_('latitude'))
    lng = models.FloatField(_('longitude'))
    alt = models.FloatField(_('altitude'), blank=True, null=True)
    status = models.CharField(_('status'), max_length=3, choices=NODE_STATUS, default='p')
    activation_key = models.CharField(_('activation key'), max_length=40, blank=True, null=True, help_text=_('Key needed for activation of the node. It\'s deleted once the node is activated.'))
    notes = models.TextField(_('notes'), blank=True, null=True)
    added = models.DateTimeField(_('added on'), auto_now_add=True)
    updated = models.DateTimeField(_('updated on'), auto_now=True)

    def get_lat(self):
        """ returns latitude as string (avoid django converting the dot . into a comma , in certain language sets) """
        return str(self.lat)

    def get_lng(self):
        """ returns longitude as string (avoid django converting the dot . into a comma , in certain language sets) """
        return str(self.lng)

    def __unicode__(self):
        return u'%s' % (self.name)

    class Meta:
        verbose_name = _('Node')
        verbose_name_plural = _('Nodes')
        db_table = 'nodeshot_node'


class OldDevice(models.Model):
    name = models.CharField(_('name'), max_length=50, unique=True)
    cname = models.SlugField(_('CNAME'), help_text=_('Name used for DNS resolution. Example: grid1 becomes grid1.nodename.domain.org. If left empty device name is used as default.'), max_length=30, blank=True, null=True)
    description = models.CharField(_('description'), max_length=255, blank=True, null=True)
    type = models.CharField(_('type'), max_length=50, blank=True, null=True)
    node = models.ForeignKey(OldNode, verbose_name=_('node'))
    routing_protocol = models.CharField(_('routing protocol'), max_length=20, choices=ROUTING_PROTOCOLS, default=DEFAULT_ROUTING_PROTOCOL)
    routing_protocol_version = models.CharField(_('routing protocol version'), max_length=10, blank=True, null=True)
    added = models.DateTimeField(_('added on'), auto_now_add=True)
    updated = models.DateTimeField(_('updated on'), auto_now=True)

    def __unicode__(self):
        return self.name

    class Meta:
        unique_together = (('node', 'cname'),)
        verbose_name = _('OldDevice')
        verbose_name_plural = _('OldDevices')
        db_table = 'nodeshot_device'


class OldHna(models.Model):
    device = models.ForeignKey(OldDevice)
    route = models.CharField(max_length=43)

    def __unicode__(self):
        return u'%s' % (self.route)

    class Meta:
        verbose_name = _('Hna')
        verbose_name_plural = _('Hna')
        db_table = 'nodeshot_hna'


class OldInterface(models.Model):
    ipv4_address = models.IPAddressField(verbose_name=_('ipv4 address'), blank=True, null=True, unique=True, default=None)
    ipv6_address = models.GenericIPAddressField(protocol='IPv6', verbose_name=_('ipv6 address'), blank=True, null=True, unique=True, default=None)
    mac_address = models.CharField(max_length=17, blank=True, null=True, unique=True, default=None)
    type = models.CharField(max_length=10, choices=INTERFACE_TYPE)
    cname = models.SlugField(_('cname'), help_text=_('Name used for DNS resolution. Example: eth0 becomes eth0.devicecname.nodename.domain.org. If left empty the interface type is used as default.'), max_length=30, blank=True, null=True)
    device = models.ForeignKey(OldDevice)
    draw_link = models.BooleanField(_('Draw links'), help_text=_('Draw links from/to this interface (not valid for VPN interfaces)'), default=True, blank=True)
    wireless_mode = models.CharField(max_length=5, choices=WIRELESS_MODE, blank=True, null=True)
    wireless_channel = models.CharField(max_length=4, choices=WIRELESS_CHANNEL, blank=True, null=True)
    wireless_polarity = models.CharField(max_length=1, choices=WIRELESS_POLARITY, blank=True, null=True)
    essid = models.CharField(max_length=50, null=True, blank=True, default=None)
    bssid = models.CharField(max_length=50, null=True, blank=True, default=None)
    status = models.CharField(_('status'), max_length=1, choices=INTERFACE_STATUS, default='u')
    added = models.DateTimeField(auto_now_add=True,)
    updated = models.DateTimeField(auto_now=True)

    def __unicode__(self):
        if self.ipv4_address:
            value = self.ipv4_address
        elif self.ipv6_address:
            value = self.ipv6_address
        elif self.mac_address:
            value = 'MAC: %s' % self.mac_address
        else:
            value = self.device.name
        return value

    class Meta:
        unique_together = (('device', 'cname'),)
        verbose_name = _('Interface')
        verbose_name_plural = _('Interfaces')
        db_table = 'nodeshot_interface'


class OldLink(models.Model):
    from_interface = models.ForeignKey(OldInterface, related_name='from_interface')
    to_interface = models.ForeignKey(OldInterface, related_name='to_interface')
    etx = models.FloatField(default=0)
    dbm = models.IntegerField(default=0)
    sync_tx = models.IntegerField(default=0)
    sync_rx = models.IntegerField(default=0)
    hide = models.BooleanField(_('Hide from map'), default=False)

    def get_quality(self, type='etx'):
        """ used to determine color of links"""
        if type == 'etx':
            if 0 < self.etx < 1.5:
               quality = 1
            elif self.etx < 3:
               quality = 2
            else:
                quality = 3
        elif type == 'dbm':
            if -83 < self.dbm < 0:
                quality = 1
            elif self.dbm > -88:
                quality = 2
            else:
                quality = 3
        return quality

    def get_etx(self):
        """ return etx as a string to avoid dot to comma conversion (it happens only with certain LANGUAGE_CODEs like IT)"""
        return str(self.etx)

    def __unicode__(self):
        return u'%s » %s' % (self.from_interface.device, self.to_interface.device)

    class Meta:
        verbose_name = _('Link')
        verbose_name_plural = _('Links')
        db_table = 'nodeshot_link'


class OldStatistic(models.Model):
    active_nodes = models.IntegerField(_('active nodes'))
    potential_nodes = models.IntegerField(_('potential nodes'))
    hotspots = models.IntegerField(_('hotspots'))
    links = models.IntegerField(_('active links'))
    km = models.FloatField(_('Km'))
    date = models.DateTimeField(_('Added on'), auto_now_add=True)

    def __unicode__(self):
        return u'%s' % (self.date)

    class Meta:
        verbose_name = _('Statistic')
        verbose_name_plural = _('Statistics')
        db_table = 'nodeshot_statistic'


class OldContact(models.Model):
    node = models.ForeignKey(OldNode)
    from_name = models.CharField(_('name'), max_length=50)
    from_email = models.EmailField(_('email'), max_length=50)
    message = models.CharField(_('message'), max_length=2000)
    ip = models.GenericIPAddressField(verbose_name=_('ip address'))
    user_agent = models.CharField(max_length=200, blank=True)
    http_referer = models.CharField(max_length=200, blank=True)
    accept_language = models.CharField(max_length=60, blank=True)
    date = models.DateTimeField(auto_now_add=True)

    def __unicode__(self):
        return _(u'Message from %(from)s to %(to)s') % ({'from':self.from_name, 'to':self.node.name})

    class Meta:
        verbose_name = _('Contact Log')
        verbose_name_plural = _('Contact Logs')
        db_table = 'nodeshot_contact'


# --- users hack --- #

from django.contrib.auth.models import AbstractBaseUser, UserManager
from django.utils import timezone


class OldUser(AbstractBaseUser):
    username = models.CharField(_('username'), max_length=30, unique=True,
        help_text=_('Required. 30 characters or fewer. Letters, numbers and '
                    '@/./+/-/_ characters'))
    first_name = models.CharField(_('first name'), max_length=30, blank=True)
    last_name = models.CharField(_('last name'), max_length=30, blank=True)
    email = models.EmailField(_('email address'), blank=True)
    is_staff = models.BooleanField(_('staff status'), default=False,
        help_text=_('Designates whether the user can log into this admin '
                    'site.'))
    is_active = models.BooleanField(_('active'), default=True,
        help_text=_('Designates whether this user should be treated as '
                    'active. Unselect this instead of deleting accounts.'))
    is_superuser = models.BooleanField(_('superuser status'), default=False,
        help_text=_('Designates that this user has all permissions without '
                    'explicitly assigning them.'))
    date_joined = models.DateTimeField(_('date joined'), default=timezone.now)

    objects = UserManager()

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = ['email']

    class Meta:
        app_label = 'oldimporter'
        db_table = 'auth_user'



from django.contrib import admin
from django.contrib.gis import forms
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

from leaflet.admin import LeafletGeoAdmin as GeoModelAdmin


class BaseAdmin(admin.ModelAdmin):
    """
    Abstract administration model for BaseDate models.
        * 'added' and 'updated' fields readonly
        * save-on-top button enabled by default
    """
    save_on_top = True
    readonly_fields = ['added', 'updated']
    html_editor_fields = []
    # preload tinymce editor static files
    if 'grappelli' in settings.INSTALLED_APPS:
        class Media:
            js = [
                '%sgrappelli/tinymce/jscripts/tiny_mce/tiny_mce.js' % settings.STATIC_URL,
                '%sgrappelli/tinymce_setup/tinymce_setup_ns.js' % settings.STATIC_URL,
            ]
        # enable editor for "node description" only
        def formfield_for_dbfield(self, db_field, **kwargs):
            field = super(BaseAdmin, self).formfield_for_dbfield(db_field, **kwargs)
            if db_field.name in self.html_editor_fields:
                field.widget.attrs['class'] = 'html-editor %s' % field.widget.attrs.get('class', '')
            return field


class BaseGeoAdmin(BaseAdmin, GeoModelAdmin):
    """
    BaseAdmin + Geodjango support
    """


class PublishActionsAdminMixin(object):
    """
    Admin mixin that adds 2 actions in the admin list:
        * publish objects
        * unpublish objects
    """
    actions = ['publish_action', 'unpublish_action']

    def publish_action(self, request, queryset):
        rows_updated = queryset.update(is_published=True)

        if rows_updated == 1:
            message_bit = _("1 item was")
        else:
            message_bit = _("%s items were") % rows_updated

        self.message_user(request, _("%s successfully published.") % message_bit)

    publish_action.short_description = _("Publish selected items")

    def unpublish_action(self, request, queryset):
        rows_updated = queryset.update(is_published=False)

        if rows_updated == 1:
            message_bit = _("1 item was")
        else:
            message_bit = _("%s items were") % rows_updated

        self.message_user(request, _("%s successfully unpublished.") % message_bit)

    unpublish_action.short_description = _("Unpublish selected items")


class BaseStackedInline(admin.StackedInline):
    readonly_fields = ['added', 'updated']
    extra = 0


class BaseTabularInline(admin.TabularInline):
    readonly_fields = ['added', 'updated']
    extra = 0



"""
nodeshot.core.nodes unit tests
"""

import os
import simplejson as json

from django.test import TestCase
from django.core.urlresolvers import reverse
from django.contrib.auth.models import AnonymousUser
from django.contrib.gis.geos import GEOSGeometry
from django.core.exceptions import ValidationError
from django.contrib.auth import get_user_model
User = get_user_model()

from nodeshot.core.base.tests import user_fixtures, BaseTestCase

from .models import Node, Status, Image


class NodeModelsTest(TestCase):
    fixtures = [
        'initial_data.json',
        user_fixtures,
        'test_layers.json',
        'test_status.json',
        'test_nodes.json',
        'test_images.json'
    ]

    def test_status_model(self):
        """ test status model internal mechanism works correctly """
        for status in Status.objects.all():
            status.delete()

        testing = Status(name='testing', slug='testing', description='slug')
        testing.save()

        self.assertEqual(testing.order, 0)
        self.assertEqual(testing.is_default, True)

        active = Status(name='active', slug='active', description='active')
        active.save()

        self.assertEqual(active.order, 1)
        self.assertEqual(active.is_default, False)

        pending = Status(name='pending', slug='pending', description='pending')
        pending.save()

        self.assertEqual(pending.order, 2)
        self.assertEqual(pending.is_default, False)

        pending.is_default = True
        pending.save()

        default_statuses = Status.objects.filter(is_default=True)
        self.assertEqual(default_statuses.count(), 1)
        self.assertEqual(default_statuses[0].pk, pending.pk)

        unconfirmed = Status(name='unconfirmed', slug='unconfirmed', description='unconfirmed', is_default=True)
        unconfirmed.save()

        default_statuses = Status.objects.filter(is_default=True)
        self.assertEqual(default_statuses.count(), 1)
        self.assertEqual(default_statuses[0].pk, unconfirmed.pk)

    def test_current_status(self):
        """ test that node._current_status is none for new nodes """
        n = Node()
        self.failUnlessEqual(n._current_status, None, 'new node _current_status private attribute is different than None')
        n = Node.objects.all().order_by('-id')[0]
        self.failUnlessEqual(n._current_status, n.status.id, 'new node _current_status private attribute is different than status')
        n.status = Status.objects.get(pk=2)
        self.failIfEqual(n._current_status, n.status.id, 'new node _current_status private attribute is still equal to status')
        n.save()
        self.failUnlessEqual(n._current_status, n.status.id, 'new node _current_status private attribute is different than status')
        n.status_id = 3
        n.save()
        n = Node.objects.all().order_by('-id')[0]
        self.failUnlessEqual(n._current_status, n.status.id, 'new node _current_status private attribute is different than status')

    def test_node_manager(self):
        """ test manager methods of Node model """
        # published()
        Node.objects.published
        count = Node.objects.published().filter(layer=1).count()
        # no unplished nodes on that layer, so the count should be the same
        self.assertEqual(count, Node.objects.filter(layer=1).count())
        # unpublish the first
        node = Node.objects.published().filter(layer=1)[0]
        node.is_published = False
        node.save()
        # should be -1
        self.assertEqual(count - 1, Node.objects.published().filter(layer=1).count())

        # Ensure GeoManager distance is available
        pnt = Node.objects.get(slug='pomezia').geometry
        Node.objects.filter(geometry__distance_lte=(pnt, 7000))

        # access level manager
        user = User.objects.get(pk=1, is_superuser=True)
        # superuser can see all nodes
        self.assertEqual(Node.objects.all().count(), Node.objects.accessible_to(user).count())
        # same but passing only user_id
        user_1 = User.objects.get(pk=1)
        self.assertEqual(Node.objects.all().count(), Node.objects.accessible_to(user_1).count())
        # simulate non authenticated user
        self.assertEqual(8, Node.objects.accessible_to(AnonymousUser()).count())
        # public nodes
        self.assertEqual(8, Node.objects.access_level_up_to('public').count())
        # public and registered
        self.assertEqual(9, Node.objects.access_level_up_to('registered').count())
        # public, registered and community
        self.assertEqual(10, Node.objects.access_level_up_to('community').count())

        # --- START CHAINING! WOOOO --- #
        # 9 because we unpublished one
        self.assertEqual(9, Node.objects.published().access_level_up_to('community').count())
        self.assertEqual(9, Node.objects.access_level_up_to('community').published().count())
        # user 1 is admin and can see all the nodes, published() is the same as writing filter(is_published=True)
        count = Node.objects.all().filter(is_published=True).count()
        self.assertEqual(count, Node.objects.published().accessible_to(user_1).count())
        self.assertEqual(count, Node.objects.accessible_to(user_1).published().count())
        # chain with geographic query
        count = Node.objects.all().filter(is_published=True).filter(layer_id=1).count()
        self.assertEqual(count, Node.objects.filter(geometry__distance_lte=(pnt, 70000)).accessible_to(user_1).published().count())
        self.assertEqual(count, Node.objects.accessible_to(user_1).filter(geometry__distance_lte=(pnt, 70000)).published().count())
        self.assertEqual(count, Node.objects.accessible_to(user_1).published().filter(geometry__distance_lte=(pnt, 70000)).count())
        self.assertEqual(count, Node.objects.filter(geometry__distance_lte=(pnt, 70000)).accessible_to(user_1).published().count())

        # slice, first, last, find
        self.assertEqual(Node.objects.last().__class__.__name__, 'Node')
        self.assertEqual(Node.objects.last(), Node.objects.order_by('-id')[0])

        self.assertEqual(Node.objects.first().__class__.__name__, 'Node')
        self.assertEqual(Node.objects.first(), Node.objects.order_by('id')[0])

        self.assertEqual(Node.objects.find(1), Node.objects.get(pk=1))

        self.assertEqual(list(Node.objects.slice('name', 5)), list(Node.objects.order_by('name')[0:5]))
        self.assertEqual(list(Node.objects.slice('-name', 5)), list(Node.objects.order_by('-name')[0:5]))

        # chained
        self.assertEqual(Node.objects.published().first(), Node.objects.filter(is_published=True).order_by('id')[0])
        self.assertEqual(Node.objects.published().last(), Node.objects.filter(is_published=True).order_by('-id')[0])

        self.assertEqual(
            Node.objects.published().access_level_up_to('public').first(),
            Node.objects.filter(is_published=True, access_level__lte=0).order_by('id')[0]
        )
        self.assertEqual(
            Node.objects.published().access_level_up_to('public').last(),
            Node.objects.filter(is_published=True, access_level__lte=0).order_by('-id')[0]
        )

    def test_autogenerate_slug(self):
        n = Node()
        n.name = 'Auto generate this'
        n.layer_id = 1
        n.geometry = 'POINT(12.509303756712 41.881163629853)'
        n.full_clean()

        n.save()

        n = Node.objects.get(pk=n.id)
        self.assertEqual(n.slug, 'auto-generate-this')

    def test_node_point(self):
        node = Node.objects.first()
        self.assertEqual(node.point, node.geometry)

        node = Node()
        with self.assertRaises(ValueError):
            node.point

        node.geometry = GEOSGeometry("""{"type": "Polygon", "coordinates": [[[12.501493164066, 41.990441051094], [12.583890625003, 41.957770034531], [12.618222900394, 41.912820024702], [12.607923217778, 41.877552973685], [12.582088180546, 41.82423212474], [12.574148841861, 41.813357913568], [12.551532455447, 41.799730560554], [12.525053688052, 41.795155470656], [12.510505386356, 41.793715689492], [12.43308610535, 41.803249638226], [12.388883300784, 41.813613798573], [12.371030517581, 41.870906276755], [12.382016845706, 41.898511105474], [12.386136718753, 41.912820024702], [12.38064355469, 41.926104006681], [12.38064355469, 41.955727539561], [12.413602539065, 41.974107637675], [12.445188232426, 41.983295698272], [12.45617456055, 41.981254021593], [12.476773925785, 41.985337309484], [12.490506835941, 41.985337309484], [12.506986328129, 41.990441051094], [12.501493164066, 41.990441051094]]]}""")
        node.point  # must not raise GEOSException

    def test_image_manager(self):
        """ test manager methods of Image model """
        # admin can see all the images
        user_1 = User.objects.get(pk=1)
        self.assertEqual(Image.objects.all().count(), Image.objects.accessible_to(user_1).count())

    def test_image_auto_order(self):
        """ test image automatic ordering works correctly """
        # node #3 has already 2 images, therefore the new image auto order should be set to 2
        image = Image(node_id=3, file='test3.jpg')
        image.full_clean()
        image.save()
        self.assertEqual(image.order, 2)

        # node #2 does not have any image, therefore the new image auto order should be set to 0
        image = Image(node_id=2, file='test2.jpg')
        image.full_clean()
        image.save()
        self.assertEqual(image.order, 0)

    def test_geometry_collection_with_single_item(self):
        node = Node.objects.get(pk=1)
        node.geometry = GEOSGeometry("GEOMETRYCOLLECTION(POINT(12.509303756712 41.881163629853))")
        node.save()
        # fetch again cos geometry value is modified while saving
        node = Node.objects.get(pk=1)

        point = GEOSGeometry("POINT(12.509303756712 41.881163629853)")
        self.assertEqual(node.geometry, point)


# ------ API tests ------ #


class NodesApiTest(BaseTestCase):

    fixtures = [
        'initial_data.json',
        user_fixtures,
        'test_layers.json',
        'test_status.json',
        'test_nodes.json',
        'test_images.json'
    ]

    def test_node_list(self):
        """ test node list """
        url = reverse('api_node_list')

        # GET: 200
        response = self.client.get(url)
        public_node_count = Node.objects.published().access_level_up_to('public').count()
        self.assertEqual(public_node_count, len(response.data['results']))

        node = {
            "layer": "rome",
            "name": "test distance",
            "address": "via dei test",
            "description": "",
            "geometry": json.loads(GEOSGeometry("POINT (12.99 41.8720419277)").json),
        }

        # POST: 403 - unauthenticated
        response = self.client.post(url, json.dumps(node), content_type='application/json')
        self.assertEqual(403, response.status_code)

        # POST: 201
        self.client.login(username='registered', password='tester')
        response = self.client.post(url, json.dumps(node), content_type='application/json')
        self.assertEqual(201, response.status_code)

        self.assertEqual(response.data['user'], 'registered')

        node = Node.objects.get(slug='test-distance')
        self.assertEqual(node.name, "test distance")

    def test_node_list_search(self):
        url = reverse('api_node_list')

        # GET: 200
        response = self.client.get(url, {"search": "Fusolab"})
        self.assertEqual(response.data['count'], 1)

    def test_node_list_filter_layers(self):
        url = reverse('api_node_list')
        response = self.client.get(url, {"layers": "rome"})
        self.assertEqual(response.data['count'], 4)
        response = self.client.get(url, {"layers": "viterbo"})
        self.assertEqual(response.data['count'], 2)
        response = self.client.get(url, {"layers": "rome,viterbo"})
        self.assertEqual(response.data['count'], 6)
        response = self.client.get(url, {"layers": "rome,viterbo,pisa"})
        self.assertEqual(response.data['count'], 8)

    def test_delete_node(self):
        node = Node.objects.first()
        node.delete()

    def test_node_geojson_list(self):
        """ test node geojson list """
        url = reverse('api_node_gejson_list')
        # GET: 200
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)

    def test_node_geojson_list_filter_layers(self):
        url = reverse('api_node_gejson_list')
        response = self.client.get(url, {"layers": "rome"})
        self.assertEqual(len(response.data['features']), 4)
        response = self.client.get(url, {"layers": "viterbo"})
        self.assertEqual(len(response.data['features']), 2)
        response = self.client.get(url, {"layers": "rome,viterbo"})
        self.assertEqual(len(response.data['features']), 6)
        response = self.client.get(url, {"layers": "rome,viterbo,pisa"})
        self.assertEqual(len(response.data['features']), 8)

    def test_node_details(self):
        """ test node details """
        url = reverse('api_node_details', args=['fusolab'])
        # GET: 200
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        node = response.data
        # images_url in node['images']
        self.assertTrue(isinstance(node['relationships']['images'], list))
        self.assertTrue(isinstance(node['relationships']['images'][0], dict))

        # PUT: 403 - must be logged in
        response = self.client.put(url)
        self.assertEqual(403, response.status_code)

        self.client.login(username='pisano', password='tester')
        # PUT: 403 - only owner can edit
        data = {
            "name": "fusolab",
            "user": "romano",
            "elev": 80.0,
            "address": "",
            "description": "Fusolab test 2",
            "access_level": "public",
            "layer": "rome",
            "geometry": json.loads(GEOSGeometry("POINT (12.582239191899999 41.872041927700003)").json)
        }
        response = self.client.put(url, json.dumps(data), content_type='application/json')
        self.assertEqual(403, response.status_code)

        self.client.logout()
        self.client.login(username='romano', password='tester')
        response = self.client.put(url, json.dumps(data), content_type='application/json')
        self.assertEqual(200, response.status_code)
        node = Node.objects.get(slug='fusolab')
        self.assertEqual(node.name, 'fusolab')
        self.assertEqual(node.description, 'Fusolab test 2')

        # PATCH
        response = self.client.patch(url, {'description': 'Patched Fusolab Desc'})
        self.assertEqual(200, response.status_code)
        node = Node.objects.get(slug='fusolab')
        self.assertEqual(node.description, 'Patched Fusolab Desc')
        self.client.logout()

        # CAN'T GET restricted if not authenticated
        fusolab = Node.objects.get(slug='fusolab')
        fusolab.access_level = 2
        fusolab.save()
        response = self.client.get(url)
        self.assertEqual(404, response.status_code)

        # Admin can get it
        self.client.login(username='admin', password='tester')
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)

        # unpublished will return 404
        fusolab.is_published = False
        fusolab.save()
        response = self.client.get(url)
        self.assertEqual(404, response.status_code)

        fusolab.is_published = True
        fusolab.save()

        # DELETE 204
        response = self.client.delete(url)
        self.assertEqual(204, response.status_code)
        with self.assertRaises(Node.DoesNotExist):
            Node.objects.get(slug='fusolab')

    def test_node_images_relationship(self):
        url = reverse('api_node_details', args=['fusolab'])

        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        images = response.data['relationships']['images']
        public_image_count = Image.objects.access_level_up_to('public').filter(node__slug='fusolab').count()
        self.assertEqual(public_image_count, len(images))
        # admin can get more images
        self.client.login(username='admin', password='tester')
        response = self.client.get(url)
        images = response.data['relationships']['images']
        node_image_count = Image.objects.accessible_to(User.objects.get(pk=1)).filter(node__slug='fusolab').count()
        self.assertEqual(node_image_count, len(images))

    def test_node_images(self):
        """ test node images """
        url = reverse('api_node_images', args=['fusolab'])

        # GET: 200
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        images = json.loads(response.content)
        public_image_count = Image.objects.access_level_up_to('public').filter(node__slug='fusolab').count()
        self.assertEqual(public_image_count, len(images))
        # admin can get more images
        self.client.login(username='admin', password='tester')
        response = self.client.get(url)
        images = json.loads(response.content)
        node_image_count = Image.objects.accessible_to(User.objects.get(pk=1)).filter(node__slug='fusolab').count()
        self.assertEqual(node_image_count, len(images))

        # GET: 404
        url = reverse('api_node_images', args=['idontexist'])
        response = self.client.get(url)
        self.assertEqual(404, response.status_code)

        # POST
        # todo
        self.client.login(username='admin', password='tester')
        good_post_data = {"description": "new image", "order": ""}
        bad_post_data = {"node": 100, "image": "jpeg99", "description": "new image", "order": ""}
        url = reverse('api_node_images', args=['fusolab'])
        wrong_url = reverse('api_node_images', args=['idontexist'])

        # wrong slug -- 404
        response = self.client.post(wrong_url, good_post_data)
        self.assertEqual(response.status_code, 404)

        # correct POST data and correct slug -- 201
        with open("%s/templates/image_unit_test.gif" % os.path.dirname(os.path.realpath(__file__)), 'rb') as image_file:
            good_post_data['file'] = image_file
            response = self.client.post(url, good_post_data)
            self.assertEqual(response.status_code, 201)
            # ensure image name is in DB
            image = Image.objects.all().order_by('-id')[0]
            self.assertIn('image_unit_test', image.file.name)
            self.assertIn('.gif', image.file.name)
            # remove file
            os.remove(image.file.file.name)

        # POST 201 - ensure additional post data "user" and "node" are ignored
        with open("%s/templates/image_unit_test.gif" % os.path.dirname(os.path.realpath(__file__)), 'rb') as image_file:
            bad_post_data['file'] = image_file
            response = self.client.post(url, bad_post_data)
            self.assertEqual(response.status_code, 201)
            image_dict = json.loads(response.content)
            self.assertEqual(image_dict['node'], 1)
            self.assertEqual(image_dict['description'], "new image")
            # ensure image name is in DB
            image = Image.objects.all().order_by('-id')[0]
            self.assertIn('image_unit_test', image.file.name)
            # remove file
            os.remove(image.file.file.name)

        self.client.logout()
        self.client.login(username='pisano', password='tester')

        with open("%s/templates/image_unit_test.gif" % os.path.dirname(os.path.realpath(__file__)), 'rb') as image_file:
            good_post_data['file'] = image_file
            response = self.client.post(url, good_post_data)
            self.assertEqual(response.status_code, 403)

    def test_node_image_list_permissions(self):
        # GET protected image should return 404
        url = reverse('api_node_images', args=['hidden-rome'])
        response = self.client.get(url)
        self.assertEqual(404, response.status_code)

        self.client.login(username='admin', password='tester')
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)

    def test_node_can_edit(self):
        url = reverse('api_node_details', args=['fusolab'])
        # cannot edit
        response = self.client.get(url)
        self.assertFalse(response.data['can_edit'])
        # admin can edit
        self.client.login(username='admin', password='tester')
        response = self.client.get(url)
        self.assertTrue(response.data['can_edit'])
        self.client.logout()
        # other user cannot edit
        self.client.login(username='registered', password='tester')
        response = self.client.get(url)
        self.assertFalse(response.data['can_edit'])
        self.client.logout()
        # owner can edit
        self.client.login(username='romano', password='tester')
        response = self.client.get(url)
        self.assertTrue(response.data['can_edit'])
        self.client.logout()
        # external node can't be edited
        Node.objects.create(**{
            'name': 'external',
            'slug': 'external',
            'layer_id': 4,
            'geometry': 'POINT (12.9732427 42.638304372)'
        })
        self.client.login(username='admin', password='tester')
        url = reverse('api_node_details', args=['external'])
        response = self.client.get(url)
        self.assertFalse(response.data['can_edit'])
        response = self.client.patch(url, {'name': 'external 2'})
        self.assertEqual(response.status_code, 403)
        self.client.logout()

    def test_node_geojson_list_pagination(self):
        # create more than 50 nodes
        for n in range(0, 60):
            Node.objects.create(
                name='node-%d' % n,
                slug='node-%d' % n,
                layer_id=1,
                geometry='POINT (42.0454 12.%d45342)' % n
            )
        url = reverse('api_node_gejson_list')
        response = self.client.get(url)
        # ensure all results are returned by default
        count = Node.objects.published().access_level_up_to('public').count()
        self.assertEqual(len(response.data['features']), count)
        # ensure pagination doesn't break geojson format
        response = self.client.get(url, {'limit': '1'})
        self.assertIn('type', response.data)
        self.assertIn('features', response.data)
        self.assertEqual(len(response.data['features']), 1)

    def test_node_integrity_error_model(self):
        node = Node(**{
            "name": "pomezia",
            "slug": "",
            "layer_id": 1,
            "geometry": 'POINT (12.50094473361969 41.66767450196442)',
            "elev": 0,
            "address": "",
            "description": ""
        })
        with self.assertRaises(ValidationError):
            node.full_clean()

    def test_node_integrity_error_api(self):
        self.client.login(username='admin', password='tester')
        url = reverse('api_node_list')
        json_data = json.dumps({
            "name": "pomezia",
            "slug": "",
            "layer": "rome",
            "geometry": {"type": "Point", "coordinates": [12.50094473361969, 41.66767450196442]},
            "elev": 0,
            "address": "",
            "description": ""
        })
        response = self.client.post(url, json_data, content_type='application/json')
        self.assertEqual(response.status_code, 400)

    def test_slug_changes_automatically(self):
        # change name 0
        existing = Node.objects.get(slug='pomezia')
        existing.name = 'Changed'
        existing.full_clean()
        existing.save()
        # query again to be sure data is up to date
        existing = Node.objects.get(pk=existing.pk)
        self.assertEqual(existing.slug, 'changed')

    def test_elevation_api(self):
        url = reverse('api_elevation_profile')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 400)
        # original response
        response = self.client.get(url, {
            'locations': '41.889040454306752,12.525333445447737',
            'original': 'true'
        })
        self.assertEqual(response.status_code, 200)
        self.assertIn('results', response.data)
        self.assertIn('status', response.data)
        # locations, geojson response - Point
        response = self.client.get(url, {
            'locations': '41.889040454306752,12.525333445447737'
        })
        self.assertEqual(response.status_code, 200)
        self.assertIn('Point', response.data['geometry']['type'])
        self.assertEqual(len(response.data['geometry']['coordinates']), 1)
        self.assertEqual(len(response.data['geometry']['coordinates'][0]), 3)
        # locations, geojson response - LineString
        response = self.client.get(url, {
            'locations': '41.889040454306752,12.525333445447737|41.889050454306752,12.525335445447737'
        })
        self.assertEqual(response.status_code, 200)
        self.assertIn('LineString', response.data['geometry']['type'])
        self.assertEqual(len(response.data['geometry']['coordinates']), 2)
        self.assertEqual(len(response.data['geometry']['coordinates'][0]), 3)
        self.assertEqual(len(response.data['geometry']['coordinates'][1]), 3)
        # path, geojson response - LineString
        response = self.client.get(url, {
            'path': '41.889040454306752,12.525333445447737|41.872041927699982,12.582239191900001',
            'samples': 4
        })
        self.assertEqual(response.status_code, 200)
        self.assertIn('LineString', response.data['geometry']['type'])
        self.assertEqual(len(response.data['geometry']['coordinates']), 4)
        self.assertEqual(len(response.data['geometry']['coordinates'][0]), 3)
        self.assertEqual(len(response.data['geometry']['coordinates'][-1]), 3)
        # path, geojson response - LineString -automatic sampling
        response = self.client.get(url, {
            'path': '41.8890404543067518,12.5253334454477372|41.8972185849048984,12.4902286938660296'
        })
        self.assertEqual(response.status_code, 200)
        self.assertIn('LineString', response.data['geometry']['type'])
        self.assertEqual(len(response.data['geometry']['coordinates']), 72)
        self.assertEqual(len(response.data['geometry']['coordinates'][0]), 3)
        self.assertEqual(len(response.data['geometry']['coordinates'][-1]), 3)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from ..settings import settings, TEXTS, USER_SETTING, DEFAULT_BOOLEAN, DEFAULT_DISTANCE


def add_notifications(myclass):
    """
    Decorator which adds fields dynamically to User Notification Settings models.

    Each of the keys in the TEXTS dictionary
    are added as a field and DB column.
    """
    for key, value in TEXTS.items():
        # custom notifications cannot be disabled
        if 'custom' in [key, value]:
            continue

        field_type = USER_SETTING[key]['type']

        if field_type == 'boolean':
            field = models.BooleanField(_(key), default=DEFAULT_BOOLEAN)
        elif field_type == 'distance':
            field = models.IntegerField(
                _(key),
                default=DEFAULT_DISTANCE,
                help_text=_('-1 (less than 0): disabled; 0: enabled for all;\
                            1 (less than 0): enabled for those in the specified distance range (km)')
            )
            field.geo_field = USER_SETTING[key]['geo_field']

        field.name = field.column = field.attname = key
        field.user_setting_type = field_type
        setattr(myclass, key, field)
        myclass.add_to_class(key, field)

    return myclass


@add_notifications
class UserWebNotificationSettings(models.Model):
    """
    User Web Notification Settings Model
    Takes care of tracking the user's web notification preferences
    (web notifications are the notifications displayed through the web interface)
    """
    user = models.OneToOneField(settings.AUTH_USER_MODEL,
                                verbose_name=_('user'),
                                related_name='web_notification_settings')

    class Meta:
        app_label = 'notifications'
        db_table = 'notifications_user_web_settings'
        verbose_name = _('user web notification settings')
        verbose_name_plural = _('user web notification settings')

    def __unicode__(self):
        return _('web notification settings for %s') % self.user


@add_notifications
class UserEmailNotificationSettings(models.Model):
    """
    User Email Notification Settings Model
    Takes care of tracking the user's email notification preferences
    """
    user = models.OneToOneField(settings.AUTH_USER_MODEL,
                                verbose_name=_('user'),
                                related_name='email_notification_settings')

    class Meta:
        app_label = 'notifications'
        db_table = 'notifications_user_email_settings'
        verbose_name = _('user email notification settings')
        verbose_name_plural = _('user email notification settings')

    def __unicode__(self):
        return _('email notification settings for %s') % self.user



from __future__ import absolute_import

import os
import celery
from django.conf import settings


__all__ = ['Celery', 'init_celery']


class Celery(celery.Celery):
    """
    send celery exceptions and stacktraces to sentry
    thanks to http://stackoverflow.com/questions/27550916/celery-tasks-uncaught-exceptions-not-being-sent-to-sentry
    """
    def on_configure(self):
        if hasattr(settings, 'RAVEN_CONFIG') and settings.RAVEN_CONFIG['dsn']:
            import raven
            from raven.contrib.celery import (register_signal,
                                              register_logger_signal)
            client = raven.Client(settings.RAVEN_CONFIG['dsn'])
            register_logger_signal(client)
            register_signal(client)


def init_celery(project_name):
    """ init celery app without the need of redundant code """
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', '%s.settings' % project_name)
    app = Celery(project_name)
    app.config_from_object('django.conf:settings')
    app.autodiscover_tasks(settings.INSTALLED_APPS, related_name='tasks')
    return app



import inspect
import time

from tld import get_tld
from tld.exceptions import TldBadUrl, TldDomainNotFound, TldIOError

from .utils import write
from . import settings


class InfluxDBRequestMiddleware(object):
    """
    Measures request time and sends metric to InfluxDB.
    Credits go to: https://github.com/andymckay/django-statsd/blob/master/django_statsd/middleware.py#L24  # NOQA
    """
    def process_view(self, request, view_func, view_args, view_kwargs):
        view = view_func
        if not inspect.isfunction(view_func):
            view = view.__class__
        try:
            request._view_module = view.__module__
            request._view_name = view.__name__
            request._start_time = time.time()
        except AttributeError:  # pragma: no cover
            pass

    def process_response(self, request, response):
        self._record_time(request)
        return response

    def process_exception(self, request, exception):
        self._record_time(request)

    def _record_time(self, request):
        if hasattr(request, '_start_time'):
            if request._view_module in settings.INFLUXDB_MIDDLEWARE_IGNORED_MODULES:
                return
            ms = int((time.time() - request._start_time) * 1000)
            if request.is_ajax():
                is_ajax = True
            else:
                is_ajax = False

            is_authenticated = False
            is_staff = False
            is_superuser = False
            if request.user.is_authenticated():
                is_authenticated = True
                if request.user.is_staff:
                    is_staff = True
                if request.user.is_superuser:
                    is_superuser = True

            referer = request.META.get('HTTP_REFERER')
            referer_tld = None
            referer_tld_string = ''
            if referer:
                try:
                    referer_tld = get_tld(referer, as_object=True)
                except (TldBadUrl, TldDomainNotFound, TldIOError):  # pragma: no cover
                    pass
            if referer_tld:
                referer_tld_string = referer_tld.tld
            # data
            values = {
                'response_time': ms,
                'is_ajax': is_ajax,
                'is_authenticated': is_authenticated,
                'is_staff': is_staff,
                'is_superuser': is_superuser,
                'referer': str(referer),
                'referer_tld': referer_tld_string,
                'full_path': request.get_full_path()
            }
            # tags, for fast retrievals
            tags = {
                'method': request.method,
                'module': request._view_module,
                'view': request._view_name,
                'path': request.path
            }
            # write into db
            write(name='http_requests', values=values, tags=tags)



from django.db.models.signals import post_syncdb
from django.contrib.auth import models

from .utils import create_database


def create_database_signal(sender, **kwargs):
    create_database()


post_syncdb.connect(create_database_signal, sender=models)



from celery import task


# ------ Asynchronous tasks ------ #


@task
def create_related_object(model, kwargs):
    """
    create object with specified kwargs in background
    """
    model.objects.create(**kwargs)


from nodeshot.core.base.utils import check_dependencies

check_dependencies(
    dependencies=[
        'nodeshot.core.nodes',
        'nodeshot.community.profiles'
    ],
    module='nodeshot.community.participation'
)

from django.core.exceptions import ObjectDoesNotExist
from django.conf import settings

from comment import Comment
from vote import Vote
from rating import Rating

from node_participation_settings import NodeParticipationSettings
from node_rating_count import NodeRatingCount


__all__ = [
    'Comment',
    'Vote',
    'Rating',
    'NodeParticipationSettings',
    'NodeRatingCount',
]


# ------ Layers Operations ------ #


if 'nodeshot.core.layers' in settings.INSTALLED_APPS:
    from layer_participation_settings import LayerParticipationSettings

    __all__ += ['LayerParticipationSettings']

    from nodeshot.core.layers.models import Layer

    @property
    def _layer_participation_settings(self):
        """
        Return layer_participation_settings record
        or create it if it does not exist

        usage:
        layer = Layer.objects.get(pk=1)
        layer.participation_settings
        """
        try:
            return self.layer_participation_settings
        except ObjectDoesNotExist:
            layer_participation_settings = LayerParticipationSettings(layer=self)
            layer_participation_settings.save()
            return layer_participation_settings

    Layer.participation_settings = _layer_participation_settings


# ------ Add methods to Node Model ------ #


from nodeshot.core.nodes.models import Node


@property
def _node_rating_count(self):
    """
    Return node_rating_count record
    or create it if it does not exist

    usage:
    node = Node.objects.get(pk=1)
    node.rating_count
    """
    try:
        return self.noderatingcount
    except ObjectDoesNotExist:
        node_rating_count = NodeRatingCount(node=self)
        node_rating_count.save()
        return node_rating_count

Node.rating_count = _node_rating_count


@property
def _node_participation_settings(self):
    """
    Return node_participation_settings record
    or create it if it does not exist

    usage:
    node = Node.objects.get(pk=1)
    node.participation_settings
    """
    try:
        return self.node_participation_settings
    except ObjectDoesNotExist:
        node_participation_settings = NodeParticipationSettings(node=self)
        node_participation_settings.save()
        return node_participation_settings

Node.participation_settings = _node_participation_settings


def _action_allowed(self, action):
    """
    participation actions can be disabled on layer level, or disabled on a per node basis
    """
    if getattr(self.layer.participation_settings, '{0}_allowed'.format(action)) is False:
        return False
    else:
        return getattr(self.participation_settings, '{0}_allowed'.format(action))


@property
def _voting_allowed(self):
    return _action_allowed(self, 'voting')


@property
def _rating_allowed(self):
    return _action_allowed(self, 'rating')


@property
def _comments_allowed(self):
    return _action_allowed(self, 'comments')


Node.voting_allowed = _voting_allowed
Node.rating_allowed = _rating_allowed
Node.comments_allowed = _comments_allowed


# ------ SIGNALS ------ #


from django.dispatch import receiver
from django.db.models.signals import post_save

from ..tasks import create_related_object


@receiver(post_save, sender=Node)
def create_node_rating_counts_settings(sender, **kwargs):
    """ create node rating count and settings"""
    created = kwargs['created']
    node = kwargs['instance']
    if created:
        # create node_rating_count and settings
        # task will be executed in background unless settings.CELERY_ALWAYS_EAGER is True
        # if CELERY_ALWAYS_EAGER is False celery worker must be running otherwise task won't be executed
        create_related_object.delay(NodeRatingCount, {'node': node})
        create_related_object.delay(NodeParticipationSettings, {'node': node})


@receiver(post_save, sender=Layer)
def create_layer_rating_settings(sender, **kwargs):
    """ create layer rating settings """
    created = kwargs['created']
    layer = kwargs['instance']
    if created:
        # create layer participation settings
        # task will be executed in background unless settings.CELERY_ALWAYS_EAGER is True
        # if CELERY_ALWAYS_EAGER is False celery worker must be running otherwise task won't be executed
        create_related_object.delay(LayerParticipationSettings, {'layer': layer})



from django.conf import settings

from nodeshot.core.base.tests import user_fixtures, BaseTestCase
from nodeshot.core.nodes.models import Node

from django.core import management


class TestWebsockets(BaseTestCase):
    """
    Test WebSockets
    """
    
    fixtures = [
        'initial_data.json',
        user_fixtures,
        'test_layers.json',
        'test_status.json',
    ]
    
    #def test_start_websocket_server(self):
    #    self.assertTrue(False, 'TODO')



import simplejson as json

from rest_framework import serializers, pagination
from rest_framework_gis import serializers as geoserializers
from rest_framework_hstore.serializers import HStoreSerializer

from nodeshot.core.base.serializers import GeoJSONPaginationSerializer
from nodeshot.core.nodes.models import Node
from nodeshot.core.nodes.serializers import NodeListSerializer

from .models import Layer
from .settings import ADDITIONAL_LAYER_FIELDS


__all__ = [
    'LayerDetailSerializer',
    'LayerListSerializer',
    'LayerNodeListSerializer',
    'GeoLayerListSerializer',
    'CustomNodeListSerializer',
    'PaginatedLayerListSerializer',
    'PaginatedGeojsonLayerListSerializer'
]


class LayerListSerializer(geoserializers.GeoModelSerializer, HStoreSerializer):
    """
    Layer list
    """
    details = serializers.HyperlinkedIdentityField(view_name='api_layer_detail', lookup_field='slug')
    nodes = serializers.HyperlinkedIdentityField(view_name='api_layer_nodes_list', lookup_field='slug')
    geojson = serializers.HyperlinkedIdentityField(view_name='api_layer_nodes_geojson', lookup_field='slug')
    center = serializers.SerializerMethodField('get_center')
    has_contact = serializers.SerializerMethodField('get_has_contact')

    def get_center(self, obj):
        return json.loads(obj.center.geojson)

    def get_has_contact(self, obj):
        return bool(obj.email)

    class Meta:
        model = Layer
        fields = [
            'id', 'slug', 'name', 'center', 'area', 'organization',
            'nodes_minimum_distance', 'new_nodes_allowed', 'is_external',
            'has_contact', 'details', 'nodes', 'geojson'
        ] + ADDITIONAL_LAYER_FIELDS


class PaginatedLayerListSerializer(pagination.PaginationSerializer):
    class Meta:
        object_serializer_class = LayerListSerializer


class PaginatedGeojsonLayerListSerializer(GeoJSONPaginationSerializer):
    class Meta:
        object_serializer_class = LayerListSerializer


class GeoLayerListSerializer(geoserializers.GeoFeatureModelSerializer, LayerListSerializer):
    class Meta:
        model = Layer
        geo_field = 'area'
        fields = ['id', 'name', 'slug'] + ADDITIONAL_LAYER_FIELDS


class LayerDetailSerializer(LayerListSerializer):
    """
    Layer details
    """
    class Meta:
        model = Layer
        fields = ['name', 'slug', 'center', 'area', 'organization', 'is_external',
                  'nodes_minimum_distance', 'new_nodes_allowed',
                  'description', 'text', 'has_contact',
                  'website', 'nodes', 'geojson'] + ADDITIONAL_LAYER_FIELDS


class CustomNodeListSerializer(NodeListSerializer):
    class Meta:
        model = Node
        fields = [
            'name', 'slug', 'user',
            'geometry', 'elev', 'address', 'description',
            'updated', 'added', 'details'
        ]
        read_only_fields = ['added', 'updated']
        geo_field = 'geometry'


class LayerNodeListSerializer(LayerDetailSerializer):
    """
    Nodes of a Layer
    """
    class Meta:
        model = Layer
        fields = ['name', 'description', 'text', 'organization', 'website'] + ADDITIONAL_LAYER_FIELDS



from django.contrib.gis.db import models
from django.contrib.gis.geos import LineString
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ValidationError

from django_hstore.fields import DictionaryField, ReferencesField

from nodeshot.core.base.managers import HStoreGeoAccessLevelManager as LinkManager
from nodeshot.core.base.models import BaseAccessLevel
from nodeshot.core.base.utils import choicify

from nodeshot.core.nodes.models import Node
from nodeshot.core.layers.models import Layer
from nodeshot.networking.net.models import Interface
from nodeshot.networking.net.models.choices import INTERFACE_TYPES

from .choices import METRIC_TYPES, LINK_STATUS, LINK_TYPES


class Link(BaseAccessLevel):
    """
    Link Model
    Designed for both wireless and wired links
    """
    type = models.SmallIntegerField(_('type'), max_length=10, null=True, blank=True,
                                    choices=choicify(LINK_TYPES), default=LINK_TYPES.get('radio'))

    # in most cases these two fields are mandatory, except for "planned" links
    interface_a = models.ForeignKey(Interface, verbose_name=_('from interface'),
                                    related_name='link_interface_from', blank=True, null=True,
                                    help_text=_('mandatory except for "planned" links (in planned links you might not have any device installed yet)'))
    interface_b = models.ForeignKey(Interface, verbose_name=_('to interface'),
                                    related_name='link_interface_to', blank=True, null=True,
                                    help_text=_('mandatory except for "planned" links (in planned links you might not have any device installed yet)'))

    # in "planned" links these two fields are necessary
    # while in all the other status they serve as a shortcut
    node_a = models.ForeignKey(Node, verbose_name=_('from node'),
                               related_name='link_node_from', blank=True, null=True,
                               help_text=_('leave blank (except for planned nodes) as it will be filled in automatically'))
    node_b = models.ForeignKey(Node, verbose_name=_('to node'),
                               related_name='link_node_to', blank=True, null=True,
                               help_text=_('leave blank (except for planned nodes) as it will be filled in automatically'))
    # shortcut
    layer = models.ForeignKey(Layer, verbose_name=_('layer'), blank=True, null=True,
                              help_text=_('leave blank - it will be filled in automatically'))

    # geospatial info
    line = models.LineStringField(blank=True, null=True, help_text=_('leave blank and the line will be drawn automatically'))

    # monitoring info
    status = models.SmallIntegerField(_('status'), choices=choicify(LINK_STATUS), default=LINK_STATUS.get('planned'))
    first_seen = models.DateTimeField(_('first time seen on'), blank=True, null=True, default=None)
    last_seen = models.DateTimeField(_('last time seen on'), blank=True, null=True, default=None)

    # technical info
    metric_type = models.CharField(_('metric type'), max_length=6,
                                   choices=choicify(METRIC_TYPES), blank=True, null=True)
    metric_value = models.FloatField(_('metric value'), blank=True, null=True)
    max_rate = models.IntegerField(_('Maximum BPS'), null=True, default=None, blank=True)
    min_rate = models.IntegerField(_('Minimum BPS'), null=True, default=None, blank=True)

    # wireless specific info
    dbm = models.IntegerField(_('dBm average'), null=True, default=None, blank=True)
    noise = models.IntegerField(_('noise average'), null=True, default=None, blank=True)

    # additional data
    data = DictionaryField(_('extra data'), null=True, blank=True,
                           help_text=_('store extra attributes in JSON string'))
    shortcuts = ReferencesField(null=True, blank=True)

    # django manager
    objects = LinkManager()

    class Meta:
        app_label = 'links'

    def __unicode__(self):
        return _(u'%s <> %s') % (self.node_a_name, self.node_b_name)

    def clean(self, *args, **kwargs):
        """
        Custom validation
            1. interface_a and interface_b mandatory except for planned links
            2. planned links should have at least node_a and node_b filled in
            3. dbm and noise fields can be filled only for radio links
            4. interface_a and interface_b must differ
            5. interface a and b type must match
        """
        if self.status != LINK_STATUS.get('planned') and (self.interface_a is None or self.interface_b is None):
            raise ValidationError(_('fields "from interface" and "to interface" are mandatory in this case'))

        if self.status == LINK_STATUS.get('planned') and (self.node_a is None or self.node_b is None):
            raise ValidationError(_('fields "from node" and "to node" are mandatory for planned links'))

        if self.type != LINK_TYPES.get('radio') and (self.dbm is not None or self.noise is not None):
            raise ValidationError(_('Only links of type "radio" can contain "dbm" and "noise" information'))

        if (self.interface_a_id == self.interface_b_id) or (self.interface_a == self.interface_b):
            raise ValidationError(_('link cannot have same "from interface" and "to interface"'))

        if (self.interface_a and self.interface_b) and self.interface_a.type != self.interface_b.type:
            format_tuple = (self.interface_a.get_type_display(), self.interface_b.get_type_display())
            raise ValidationError(_('link cannot be between of interfaces of different types:\
                                    interface a is "%s" while b is "%s"') % format_tuple)

    def save(self, *args, **kwargs):
        """
        Custom save does the following:
            * determine link type if not specified
            * automatically fill 'node_a' and 'node_b' fields if necessary
            * draw line between two nodes
            * fill shortcut properties node_a_name and node_b_name
        """
        if not self.type:
            if self.interface_a.type == INTERFACE_TYPES.get('wireless'):
                self.type = LINK_TYPES.get('radio')
            elif self.interface_a.type == INTERFACE_TYPES.get('ethernet'):
                self.type = LINK_TYPES.get('ethernet')
            else:
                self.type = LINK_TYPES.get('virtual')

        if self.interface_a_id:
            self.interface_a = Interface.objects.get(pk=self.interface_a_id)
        if self.interface_b_id:
            self.interface_b = Interface.objects.get(pk=self.interface_b_id)

        # fill in node_a and node_b
        if self.node_a is None and self.interface_a is not None:
            self.node_a = self.interface_a.node
        if self.node_b is None and self.interface_b is not None:
            self.node_b = self.interface_b.node

        # fill layer from node_a
        if self.layer is None:
            self.layer = self.node_a.layer

        # draw linestring
        if not self.line:
            self.line = LineString(self.node_a.point, self.node_b.point)

        # fill properties
        if self.data is None or self.data.get('node_a_name', None) is None:
            self.data = self.data or {}  # in case is None init empty dict
            self.data['node_a_name'] = self.node_a.name
            self.data['node_b_name'] = self.node_b.name

        if self.data.get('node_a_slug', None) is None or self.data.get('node_b_slug', None) is None:
            self.data['node_a_slug'] = self.node_a.slug
            self.data['node_b_slug'] = self.node_b.slug

        if self.data.get('interface_a_mac', None) is None or self.data.get('interface_b_mac', None) is None:
            self.data['interface_a_mac'] = self.interface_a.mac
            self.data['interface_b_mac'] = self.interface_b.mac

        if self.data.get('layer_slug') != self.layer.slug:
            self.data['layer_slug'] = self.layer.slug

        super(Link, self).save(*args, **kwargs)

    @property
    def node_a_name(self):
        self.data = self.data or {}
        return self.data.get('node_a_name', None)

    @property
    def node_b_name(self):
        self.data = self.data or {}
        return self.data.get('node_b_name', None)

    @property
    def node_a_slug(self):
        self.data = self.data or {}
        return self.data.get('node_a_slug', None)

    @property
    def node_b_slug(self):
        self.data = self.data or {}
        return self.data.get('node_b_slug', None)

    @property
    def interface_a_mac(self):
        self.data = self.data or {}
        return self.data.get('interface_a_mac', None)

    @property
    def interface_b_mac(self):
        self.data = self.data or {}
        return self.data.get('interface_b_mac', None)

    @property
    def layer_slug(self):
        self.data = self.data or {}
        return self.data.get('layer_slug', None)

    @property
    def quality(self):
        """
        Quality is a number between 1 and 6 that rates the quality of the link.
        The way quality is calculated might be overridden by settings.
        0 means unknown
        """
        if self.metric_value is None:
            return 0
        # PLACEHOLDER
        return 6



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.utils.translation import ugettext_lazy as _
from django.contrib import messages

from django.contrib.auth.forms import UserChangeForm as BaseChangeForm
from django.contrib.auth.forms import UserCreationForm as BaseCreationForm
from django.contrib.auth.forms import AdminPasswordChangeForm as BasePasswordChangeForm
from django.forms import ValidationError

from nodeshot.core.base.admin import BaseStackedInline
from .models import SocialLink, Profile, PasswordReset
from .settings import settings, EMAIL_CONFIRMATION


# --- User management forms --- #


class UserChangeForm(BaseChangeForm):
    class Meta:
        model = Profile


class UserCreationForm(BaseCreationForm):
    # this happens to be needed, at least in django 1.6.2
    # http://stackoverflow.com/questions/16953302/django-custom-user-model-in-admin-relation-auth-user-does-not-exist
    def clean_username(self):
        username = self.cleaned_data["username"]
        try:
            Profile.objects.get(username=username)
        except Profile.DoesNotExist:
            return username
        raise ValidationError(self.error_messages['duplicate_username'])

    class Meta:
        model = Profile


class AdminPasswordChangeForm(BasePasswordChangeForm):
    class Meta:
        model = Profile


# --- Admin --- #


class ProfileSocialLinksInline(BaseStackedInline):
    model = SocialLink
    extra = 0

USER_ADMIN_INLINES = [ProfileSocialLinksInline]


if 'social_auth' in settings.INSTALLED_APPS:
    from social_auth.models import UserSocialAuth

    class SocialAuthInline(admin.StackedInline):
        model = UserSocialAuth
        extra = 0

    USER_ADMIN_INLINES.append(SocialAuthInline)


class UserAdmin(BaseUserAdmin):
    list_display = (
        'username',
        'email',
        'first_name',
        'last_name',
        'is_active',
        'date_joined',
        'last_login',
        'is_staff',
        'is_superuser'
    )
    inlines = USER_ADMIN_INLINES
    ordering = ['-is_staff', '-date_joined']
    search_fields = ('email', 'username', 'first_name', 'last_name')
    list_filter = ('is_active', 'is_staff', 'is_superuser')
    form = UserChangeForm
    add_form = UserCreationForm
    change_password_form = AdminPasswordChangeForm

    fieldsets = [
        [None, {'fields': ('username', 'password')}],
        [_('Personal info'), {'fields': [
            'first_name', 'last_name', 'email',
            'birth_date', 'address', 'city', 'country',
            'gender', 'about'
        ]}],
        [_('Permissions'), {'fields': ('is_active', 'is_staff', 'is_superuser',
                                       'groups', 'user_permissions')}],
        [_('Important dates'), {'fields': ('last_login', 'date_joined')}],
    ]


class PasswordResetAdmin(admin.ModelAdmin):
    pass
    list_display = ('user', 'timestamp', 'reset', 'temp_key')


admin.site.register(Profile, UserAdmin)
admin.site.register(PasswordReset, PasswordResetAdmin)


if EMAIL_CONFIRMATION:
    from .models import EmailAddress, EmailConfirmation

    class EmailAddressAdmin(admin.ModelAdmin):
        search_fields = ('email', 'user__username')
        list_select_related = True
        list_display = ('__unicode__', 'verified', 'primary', 'user')

    class EmailConfirmationAdmin(admin.ModelAdmin):
        list_display = ('__unicode__', 'key_expired', 'created_at')
        actions = ['resend']

        def resend(self, request, queryset):
            """
            Resend confirmation mail
            """
            for obj in queryset:
                EmailConfirmation.objects.send_confirmation(obj.email_address)
            # show message in the admin
            messages.info(request, _('Email confirmations sent'), fail_silently=True)
        resend.short_description = _('Resend confirmation mails')

    admin.site.register((EmailAddress,), EmailAddressAdmin)
    admin.site.register((EmailConfirmation,), EmailConfirmationAdmin)

    class EmailAddressInline(admin.StackedInline):
        model = EmailAddress
        extra = 0

    UserAdmin.inlines = [EmailAddressInline] + UserAdmin.inlines



import datetime
from random import random
from hashlib import sha1

from django.db import models, IntegrityError
from django.core.mail import send_mail
from django.core.urlresolvers import reverse, NoReverseMatch
from django.core.exceptions import ValidationError
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _
from django.utils import timezone

from ..signals import email_confirmed, email_confirmation_sent
from ..settings import settings

try:
    from django.utils.timezone import now
except ImportError:
    now = datetime.datetime.now


__all__ = [
    'EmailAddress',
    'EmailConfirmation'
]


class EmailAddressManager(models.Manager):
    def add_email(self, user, email):
        try:
            email_address = self.create(user=user, email=email)
            EmailConfirmation.objects.send_confirmation(email_address)
            return email_address
        except IntegrityError:
            return None

    def get_primary(self, user=None):
        # if we're a related_manager, use the User object we're attached to
        inst = getattr(self, 'instance', None)
        if not user and not inst:
            raise ValueError("Must specify a User")
        if user and inst and user != inst:
            raise ValueError("Cannot specify a different User")
        user = user or inst

        try:
            return self.get(user=user, primary=True)
        except EmailAddress.DoesNotExist:
            return None

    def get_users_for(self, email):
        """
        returns a list of users with the given email.
        """
        # this is a list rather than a generator because we probably want to
        # do a len() on it right away
        return [address.user for address in EmailAddress.objects.filter(verified=True, email=email)]


class EmailAddress(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='email_set')
    email = models.EmailField(unique=True)
    verified = models.BooleanField(default=False)
    primary = models.BooleanField(default=False)

    objects = EmailAddressManager()

    def clean(self):
        try:
            if self.pk and not self.primary and self.user.email_set.get_primary().pk == self.pk:
                raise ValidationError(_("You must have at least one primary email address."))
        except IndexError:
            pass

    def save(self, *args, **kwargs):
        set_as_primary = False
        # if primary is True and this is not the only primary email address for this user
        if self.primary and self.user.email_set.filter(primary=True).exclude(pk=self.pk).count() > 0:
            set_as_primary = True
        # if primary is not true but no primary email addresses for this user
        if not self.primary and self.user.email_set.filter(primary=True).count() < 1:
            set_as_primary = True
        # if adding a new primary address from the admin
        if not self.pk and self.primary:
            set_as_primary = True
        super(EmailAddress, self).save(*args, **kwargs)
        if set_as_primary:
            # ensure there's only 1 primary address
            self.set_as_primary()

    def set_as_primary(self):
        old_primaries = self.user.email_set.filter(primary=True).exclude(pk=self.pk)
        for old_primary in old_primaries:
            old_primary.primary = False
            old_primary.save()
        self.primary = True
        self.save()

        self.user.email = self.email
        update_fields = ['email']
        if getattr(settings, 'EMAIL_CONFIRMATION_OVERWRITE_USERNAME', False):
            self.user.username = self.email
            update_fields.append('username')
        try:
            # django 1.5+
            # necessary because of more aggresive related model caching
            self.user.save(update_fields=update_fields)
        except TypeError:
            self.user.save()

        return True

    def __unicode__(self):
        return u"%s (%s)" % (self.email, self.user)

    class Meta:
        ordering = ['-primary', 'id']
        app_label = 'profiles'
        verbose_name = _("email address")
        verbose_name_plural = _("email addresses")
        unique_together = (
            ("user", "email"),
        )


class EmailConfirmationManager(models.Manager):
    def generate_key(self, email):
        """
        Generate a new email confirmation key and return it.
        """
        salt = sha1(str(random())).hexdigest()[:5]
        return sha1(salt + email).hexdigest()

    def create_emailconfirmation(self, email_address):
        "Create an email confirmation obj from the given email address obj"
        confirmation_key = self.generate_key(email_address.email)
        confirmation = self.create(
            email_address=email_address,
            created_at=now(),
            key=confirmation_key,
        )
        return confirmation

    def confirm_email(self, key, user=None, make_primary=True):
        queryset = self.all()
        if user:
            queryset = queryset.filter(email_address__user=user)
        try:
            confirmation = queryset.get(key=key)
        except self.model.DoesNotExist:
            return None
        if not confirmation.key_expired():
            email_address = confirmation.email_address
            email_address.verified = True
            if make_primary:
                email_address.set_as_primary()
            email_address.save()
            email_confirmed.send(sender=self.model, email_address=email_address)
            return email_address

    def send_confirmation(self, email_address):
        confirmation = self.create_emailconfirmation(email_address)
        # check for the url with the dotted view path
        try:
            path = reverse("emailconfirmation.views.confirm_email", args=[confirmation.key])
        except NoReverseMatch:
            # or get path with named urlconf instead
            path = reverse("emailconfirmation_confirm_email", args=[confirmation.key])
        activate_url = u"%s%s" % (settings.SITE_URL, path)
        context = {
            "user": email_address.user,
            "activate_url": activate_url,
            "site_name": settings.SITE_NAME,
            "confirmation_key": confirmation.key,
        }
        subject = render_to_string("profiles/email_messages/email_confirmation_subject.txt", context)
        # remove superfluous line breaks
        subject = "".join(subject.splitlines())
        message = render_to_string("profiles/email_messages/email_confirmation_message.txt", context)
        send_mail(subject, message, settings.DEFAULT_FROM_EMAIL, [email_address.email])

        email_confirmation_sent.send(sender=self.model, confirmation=confirmation)
        return confirmation

    def delete_expired_confirmations(self):
        for confirmation in self.all():
            if confirmation.key_expired():
                confirmation.delete()


class EmailConfirmation(models.Model):
    email_address = models.ForeignKey(EmailAddress)
    created_at = models.DateTimeField()
    key = models.CharField(max_length=40)

    objects = EmailConfirmationManager()

    def key_expired(self):
        confirmation_days = getattr(settings, 'EMAIL_CONFIRMATION_DAYS', 7)
        expiration_date = self.created_at + datetime.timedelta(days=confirmation_days)
        return expiration_date <= timezone.now()
    key_expired.boolean = True

    def __unicode__(self):
        return u"confirmation for %s" % self.email_address

    class Meta:
        app_label = 'profiles'
        verbose_name = _("email confirmation")
        verbose_name_plural = _("email confirmations")
